"""Exemple module with simple convlution neural network"""
import torch
import torch.nn as nn

from kit4dl import Kit4DLAbstractModule


class SimpleCNN(Kit4DLAbstractModule):
    def configure(self, input_dims, layers, dropout, output_dims):
        self.l1 = nn.Sequential(
            nn.Conv2d(input_dims, 16, kernel_size=3, bias=True),
            nn.ReLU(),
            nn.BatchNorm2d(16),
        )

        hidden_layers = []
        for _ in range(layers):
            hidden_layers.extend(
                [
                    nn.Conv2d(16, 16, kernel_size=3, bias=True),
                    nn.ReLU(),
                    nn.BatchNorm2d(16),
                ]
            )
        self.hidden_layers = nn.Sequential(*hidden_layers)
        self.fc = nn.Sequential(
            nn.Flatten(),
            nn.Linear(16 * 5, 64),
            nn.ReLU(),
            nn.Dropout(dropout),
            nn.Linear(64, output_dims),
        )

    def run_step(self, batch, batch_idx) -> tuple[torch.Tensor, torch.Tensor]:
        x, label = batch
        x = self.l1(x)
        x = self.hidden_layers(x)
        return label, self.fc(x)



import os

import pytest

try:
    import tomllib as toml
except ModuleNotFoundError:
    import toml

from unittest.mock import MagicMock, Mock, PropertyMock

import torch
import torchmetrics as tm

from kit4dl.nn.confmodels import Conf


@pytest.fixture
def conf():
    conf = MagicMock()
    conf.base = PropertyMock()
    conf.base.accelerator_device_and_id = PropertyMock(
        return_value="cuda"
    ), PropertyMock(return_value=0)
    conf.logging.level = "INFO"
    conf.logging.format_ = (
        "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    )
    conf.base.seed = 0
    conf.model = PropertyMock()
    conf.model.arguments = {"input_dims": 10, "output_dims": 1}
    conf.metrics_obj = {
        "Precision": tm.Precision(task="binary"),
        "FBetaScore": tm.FBetaScore(task="binary"),
    }
    conf.training = PropertyMock()
    conf.training.optimizer = PropertyMock()
    conf.training.optimizer.optimizer = torch.optim.SGD
    conf.training.checkpoint.mode = "max"

    conf.model.model_class = PropertyMock()

    conf.dataset = PropertyMock()
    conf.dataset.datamodule_class = PropertyMock()

    conf.dict = Mock()
    conf.dict.return_value = {}
    yield conf


@pytest.fixture
def true_conf():
    yield Conf(
        root_dir=os.path.join(os.getcwd(), "tests", "resources"),
        **toml.load(os.path.join("tests", "resources", "test_file.toml")),
    )


@pytest.fixture
def base_conf_txt():
    return """
    [base]
    seed = 0
    experiment_name = "handwritten_digit_classification"

    [model]
    target = "./tests/dummy_module.py::B"
    input_dims = 1
    layers = 4
    dropout = 0.5
    output_dims = 10

    [training]
    epochs = 100
    epoch_schedulers = [
        {target = "torch.optim.lr_scheduler::CosineAnnealingLR", T_max = 100}
    ]

    [training.checkpoint]
    path = "chckpt"
    monitor = {"metric" = "Precision", "stage" = "val"}
    filename = "{epoch}_{val_fbeta_score:.2f}_convlstm"
    mode = "max"

    [training.optimizer]
    target = "torch.optim::Adam"
    lr = 0.0001
    weight_decay = 0.03

    [training.criterion]
    target = "torch.nn::NLLLoss"
    weight = [0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1]

    [validation]
    run_every_epoch = 1

    [dataset]
    target = "./tests/dummy_module.py::DummyDatasetModule"

    [dataset.trainval]
    root_dir = "./mnist"

    [dataset.train.loader]
    batch_size = 10
    shuffle = true
    num_workers = 4

    [dataset.validation.loader]
    batch_size = 10
    shuffle = false
    num_workers = 4
    """


@pytest.fixture
def base_conf_txt_full(base_conf_txt):
    return base_conf_txt + """
        [metrics]
        Precision = {target = "torchmetrics::Precision", task = "multiclass", num_classes = 10}
    """


@pytest.fixture
def dummy_optimizer():
    yield torch.optim.Adam([torch.nn.Parameter(torch.rand((100,)))], lr=0.1)



"""Docstring of the module containing my neural network module."""
from typing import Any

import torch

from kit4dl import Kit4DLAbstractModule


class MyNewNetwork(Kit4DLAbstractModule):
    """My network module."""

    def configure(self, *args: Any, **kwargs: Any) -> None:
        """Configure the architecture."""
        # You can define kwargs used in the configuration file
        # def setup(self, input_dims, dropout_rate):
        #     ...
        # Here, you define the architecture
        raise NotImplementedError

    def run_step(self, batch, batch_idx) -> tuple[torch.Tensor, torch.Tensor]:
        """Run single step."""
        raise NotImplementedError



"""Module with stages utils."""
from enum import Enum


class Stage(Enum):
    """ML procedure stages."""

    TRAIN = "train"
    VALIDATION = "val"
    TEST = "test"
    PREDICT = "predict"

    def __str__(self) -> str:
        """Get text representation fo the `Stage`."""
        return self.value



import pytest


class TestLoggingMixin:
    @pytest.mark.parametrize(
        "log_method", ["debug", "info", "warn", "error", "critical"]
    )
    def test_log_methods_are_available(self, log_method):
        from kit4dl.mixins import LoggerMixin

        class A(LoggerMixin):
            pass

        assert hasattr(A(), log_method)



"""A module with resuable validators."""
import torch

from kit4dl.io import import_and_get_attr_from_fully_qualified_name
from kit4dl.kit4dl_types import FullyQualifiedName


def validate_cuda_device_exists(cuda_id: int | None = None) -> int | None:
    """Assert the indicated CUDA device exists.

    Parameters
    ----------
    cuda_id : optional int
        ID of the CUDA device. `None` means CPU device

    Raises
    ------
    AssertionError
        if `cuda_id` is not `None` and CUDA is not available
        if `cuda_id` is not `None` and the requested device does not exist
    """
    if cuda_id is None:
        return None
    assert torch.cuda.is_available(), "CUDA is not available"
    assert (
        cuda_id < torch.cuda.device_count()
    ), f"CUDA device with id `{cuda_id}` does not exist"
    return cuda_id


def validate_class_exists(
    path: FullyQualifiedName | str,
) -> FullyQualifiedName | str:
    """Validate the class defined by `path` exists.

    Parameters
    ----------
    path : str or FullyQualifiedName
        Path to the class to verify

    Returns
    -------
    path : str or FullyQualifiedName
        The path passed as an argument

    Raises
    ------
    ValueError
        if the path does not contain class separator `::`
    ModuleNotFoundError
        if module was not found
    """
    _ = import_and_get_attr_from_fully_qualified_name(path)
    return path



"""A module with input-output operations."""
import importlib
import os
import sys
from typing import Any

from kit4dl.kit4dl_types import FullyQualifiedName

_TARGET_SPLIT = "::"
PROJECT_DIR = "PROJECT_DIR"


def split_target(target: str | FullyQualifiedName) -> tuple[str, str]:
    """Split target name or fully qualified name by `::` to get the path and the attribute.

    Parameters
    ----------
    target : str or FullyQualifiedName
        The value of the class or Python module path

    Returns
    -------
    path_attr : tuple of str
        A tuple whose first item is a path component of the target and the second - attribute name

    Example
    -------
    ```
    >>> split_target("a.b.c::D")
    "a.b.c", "D"
    ```
    """
    path, attr_name = target.rsplit(_TARGET_SPLIT, maxsplit=1)
    return (path, attr_name)


def connect_target(path: str, attr_name: str) -> str:
    """Connect path and attribute name with `::`.

    Parameters
    ----------
    path : str
        The path
    attr_name: str
        The attribute name

    Returns
    -------
    target : str
        A string of `path` and `attr_name` connected with `::`

    Example
    -------
    ```
    >>> connect_target("a.b.c", "D")
    "a.b.c::D"
    ```
    """
    return _TARGET_SPLIT.join([path, attr_name])


def maybe_get_abs_target(
    target: str | FullyQualifiedName, root_dir: str
) -> str:
    """Get absolute path for the target.

    If `target` provided as the absolut path or as the fully qualified name
    of a module, no changes is done and `target` arguments is returned
    directly. If `target` is a relative path to a Python module, it is
    converted to the absolute path-based target.

    Parameters
    ----------
    target : str or FullyQualifiedName
        The value of the class or Python module path
    root_dir : str
        Root dir with respect to which absolut path should be computed

    Returns
    -------
    abs_path : str
        Absolute path for the provided target

    Example
    -------
    ```
    >>> maybe_get_abs_target("a.py::MyClass", "/work/usr")
    "/work/usr/a.py::MyClass"
    ```

    ```
    >>> maybe_get_abs_target("os::PathLike", "/work/usr")
    os::PathLike
    ```

    """
    if ".py" not in target:
        return target
    path, attr_name = split_target(target)
    if os.path.isabs(path):
        return target
    return connect_target(os.path.join(root_dir, path), attr_name)


def import_module_from_file(path: str, exec_module: bool = False):
    """Import module from file indicated by the relative path."""
    assert ".py" in path, f"path: {path} is not a Python module"
    assert os.path.exists(path), f"module: {path} does not exist"
    spec = importlib.util.spec_from_file_location(
        "_file_imported_module", path
    )
    if spec is None:
        raise RuntimeError(f"module {path} is not defined")
    _file_imported_module = importlib.util.module_from_spec(spec)
    if exec_module:
        spec.loader.exec_module(_file_imported_module)  # type: ignore[union-attr]
    return _file_imported_module


def get_class_from_py_file(path: str, name: str):
    """Get class defined in the Python file.

    Parameters
    ----------
    path : str
        The path to the Python file where the class is defined
    name : str
        The name of the class defined in the Python module indicated by the `path` argument

    Returns
    -------
    clss : type
        A class loaded from the Python file

    Example
    -------
    ```
    >>> et_class_from_py_file("./my_module.py", "MyClass")
    <class '_file_imported_module.MyClass'>
    ```
    """
    _file_imported_module = import_module_from_file(
        path=path, exec_module=True
    )
    sys.modules["_file_imported_module"] = _file_imported_module
    return getattr(_file_imported_module, name)


def import_and_get_attr_from_fully_qualified_name(
    name: FullyQualifiedName | str,
) -> Any:
    """Take class based on fully qualified name or module path.

    Parameters
    ----------
    name :  FullyQualifiedName or str
        Fully qualified name to the class, i.e. my_package.my_module::MyClass
        or path (absolute or relative) to the module together with class name:
        my_module.py::MyClass

    Returns
    -------
    res_class : type
        A type from the available module or imported from a .py file

    Raises
    ------
    ValueError
        if `name` format is wrong (does not contain class separator `::`)
    ModuleNotFoundError
        if a module (installed or from file) was not found

    Examples
    --------
    ```bash
    >>> import_and_get_attr_from_fully_qualified_name("sklearn.ensemble::RandomForestClassifier")
    <class 'sklearn.ensemble._forest.RandomForestClassifier'>
    ```

    ```bash
    >>> import_and_get_attr_from_fully_qualified_name("./dataset::MyMNISTDatamodule")
    <class '_file_imported_module.MyMNISTDatamodule'>
    ```
    """
    assert name is not None, "`name` argument cannot be `None`"
    if "::" not in name:
        raise ValueError(
            f"missing `::` in the name: `{name}`. remember to put the double"
            " colon `::` between: \n a) the fully qualified name and the"
            " class, like `package.module::MyClass` \n 2) the module path and"
            " the class, like `/usr/my_user/my_module.py::MyClass`"
        )
    path, attr_name = split_target(name)
    if ".py" in path:
        return get_class_from_py_file(path, attr_name)
    if importlib.util.find_spec(path):
        return getattr(importlib.import_module(path), attr_name)
    raise ModuleNotFoundError(
        f"module defined as `{path}` cannot be imported or found in the system"
    )



import importlib
import sys

import pytest

try:
    import tomllib as toml
except ModuleNotFoundError:
    import toml

import lightning.pytorch.loggers as pl_logs
import torch
import torchmetrics as tm
from pydantic import ValidationError

from kit4dl.nn.base import Kit4DLAbstractModule
from kit4dl.nn.confmodels import (
    BaseConf,
    CheckpointConf,
    Conf,
    CriterionConf,
    DatasetConf,
    LoggingConf,
    ModelConf,
    OptimizerConf,
    _AbstractClassWithArgumentsConf,
)
from tests.fixtures import (
    base_conf_txt,
    base_conf_txt_full,
    dummy_optimizer,
    true_conf,
)
from tests.utils import skipnocuda


class TestBaseConfAndAccessor:
    @skipnocuda
    def test_base_conf_all_passed(self):
        load = """
    seed = 0
    cuda_id = 0
    experiment_name = "handwritten_digit_classification"
        """
        conf = BaseConf(**toml.loads(load))
        assert conf.seed == 0
        assert conf.cuda_id == 0
        assert conf.experiment_name == "handwritten_digit_classification"

    def test_base_conf_no_cuda_id_passed(self):
        load = """
    seed = 0
    experiment_name = "handwritten_digit_classification"
        """
        conf = BaseConf(**toml.loads(load))
        assert conf.seed == 0
        assert conf.cuda_id is None
        assert conf.experiment_name == "handwritten_digit_classification"

    def test_base_conf_no_seed_passed(self):
        load = """
    experiment_name = "handwritten_digit_classification"
        """
        conf = BaseConf(**toml.loads(load))
        assert conf.seed == 0
        assert conf.cuda_id is None
        assert conf.experiment_name == "handwritten_digit_classification"

    def test_base_conf_failed_on_missing_exp_name(self):
        load = """
    seed = 10
    cuda_id = 1
        """
        with pytest.raises(ValidationError):
            _ = BaseConf(**toml.loads(load))

    def test_base_conf_failed_on_wrong_seed(self):
        load = """
    seed_id = -10
        """
        with pytest.raises(ValidationError):
            _ = BaseConf(**toml.loads(load))

    def test_base_conf_failed_on_wrong_cuda_id(self):
        load = """
    cuda_id = -1
        """
        with pytest.raises(ValidationError):
            _ = BaseConf(**toml.loads(load))

        load = """
    cuda_id = 10
        """
        with pytest.raises(ValidationError):
            _ = BaseConf(**toml.loads(load))

    @skipnocuda
    def test_get_cuda_device_with_id_0(self):
        load = """
    cuda_id = 0
    experiment_name = "handwritten_digit_classification"
        """
        dev = BaseConf(**toml.loads(load)).device
        assert dev.type == "cuda"
        assert dev.index == 0

    def test_get_cpu_device(self):
        load = """
    experiment_name = "handwritten_digit_classification"
        """
        dev = BaseConf(**toml.loads(load)).device
        assert dev.type == "cpu"
        assert dev.index is None


@pytest.mark.skipif(
    importlib.util.find_spec("torch") is None,
    reason="`torch` module is not installed",
)
class TestBaseClassWithArgumentsConf:
    def test_installed_class_exists(self):
        load = """
            target = "torch.optim::SGD"
        """
        conf = _AbstractClassWithArgumentsConf(**toml.loads(load))
        assert conf.target == "torch.optim::SGD"

    def test_customed_class_exists_on_path(self):
        load = """
            target = "tests.dummy_module::A"
        """
        conf = _AbstractClassWithArgumentsConf(**toml.loads(load))
        assert conf.target == "tests.dummy_module::A"

    def test_class_with_arguments(self):
        load = """
            target = "torch.optim::SGD"
            arg1 = "val1"
            arg2 = -1.5
        """
        conf = _AbstractClassWithArgumentsConf(**toml.loads(load))
        assert "arg1" in conf.arguments
        assert conf.arguments["arg1"] == "val1"
        assert "arg2" in conf.arguments
        assert conf.arguments["arg2"] == -1.5


class TestModelConf:
    def test_get_model_proper_parent_class(self):
        load = """
            target = "tests.dummy_module::B"
            input_dims = 1
            layers = 4
            dropout = 0.1
            output_dims = 10
        """
        conf = ModelConf(**toml.loads(load))
        assert issubclass(conf.model_class, Kit4DLAbstractModule)

    def test_get_model_failed_on_wrong_parent_class(self):
        load = """
            target = "tests.dummy_module::A"
        """
        with pytest.raises(ValidationError):
            _ = ModelConf(**toml.loads(load))


class TestOptimizerConf:
    def test_get_optimizer_proper_parent_class(self):
        load = """
            target = "torch.optim::SGD"
            lr = 0.1
        """
        conf = OptimizerConf(**toml.loads(load))
        assert callable(conf.optimizer)

        class _A(torch.nn.Module):
            def __init__(self):
                super().__init__()
                self.l1 = torch.nn.Linear(1, 1)

        optimizer = conf.optimizer(_A().parameters())
        assert isinstance(optimizer, torch.optim.Optimizer)

    def test_get_optimizer_failed_on_missing_lr(self):
        load = """
            target = "torch.optim::SGD"
        """
        with pytest.raises(ValidationError):
            _ = OptimizerConf(**toml.loads(load))

    def test_get_optimizer_failed_on_wrong_parent_class(self):
        load = """
            target = "tests.dummy_module::A"
            lr = 0.1
        """
        with pytest.raises(ValidationError):
            _ = OptimizerConf(**toml.loads(load))


class TestCheckpointConf:
    def test_checkpoint_conf(self):
        load = """
            path = "chckpt"
            monitor = {"metric" = "Precision", "stage" = "val"}
            filename = "{epoch}_{val_fbeta_score:.2f}_convlstm"
            mode = "max"  
        """
        conf = CheckpointConf(**toml.loads(load))
        assert conf.path == "chckpt"
        assert conf.monitor == {"metric": "Precision", "stage": "val"}
        assert conf.filename == "{epoch}_{val_fbeta_score:.2f}_convlstm"
        assert conf.mode == "max"

    def test_checkpoint_missing_metric_monitor(self):
        load = """
            path = "chckpt"
            monitor = {"stage" = "val"}
            filename = "{epoch}_{val_fbeta_score:.2f}_convlstm"
            mode = "max"  
        """
        with pytest.raises(ValidationError):
            _ = CheckpointConf(**toml.loads(load))

    def test_checkpoint_missing_stage_monitor(self):
        load = """
            path = "chckpt"
            monitor = {"metric" = "Precision"}
            filename = "{epoch}_{val_fbeta_score:.2f}_convlstm"
            mode = "max"  
        """
        with pytest.raises(ValidationError):
            _ = CheckpointConf(**toml.loads(load))

    def test_checkpoint_wrong_mode(self):
        load = """
            path = "chckpt"
            monitor = {"metric" = "Precision", "stage" = "val"}
            filename = "{epoch}_{val_fbeta_score:.2f}_convlstm"
            mode = "average"  
        """
        with pytest.raises(ValidationError):
            _ = CheckpointConf(**toml.loads(load))

        load = """
            path = "chckpt"
            monitor = {"metric" = "Precision", "stage" = "val"}
            filename = "{epoch}_{val_fbeta_score:.2f}_convlstm"
            mode = 0
        """
        with pytest.raises(ValidationError):
            _ = CheckpointConf(**toml.loads(load))

    def test_checkpoint_monitor_metric(self):
        load = """
            path = "chckpt"
            monitor = {"metric" = "Precision", "stage" = "val"}
            filename = "{epoch}_{val_fbeta_score:.2f}_convlstm"
            mode = "max"  
        """
        conf = CheckpointConf(**toml.loads(load))
        assert conf.monitor_metric_name == "val_precision"


class TestCriterionConfig:
    def test_criterion(self):
        load = """
            target = "torch.nn::NLLLoss"
        """
        conf = CriterionConf(**toml.loads(load))
        conf.target == "torch.nn.NLLLoss"

    def test_criterion_failed_on_weight_passed_if_not_supported(self):
        load = """
            target = "torch.nn::MSELoss"
            weight = [0.1, 0.1]
        """
        with pytest.raises(ValidationError):
            _ = CriterionConf(**toml.loads(load))

    def test_criterion_weight_passed(self):
        load = """
            target = "torch.nn::NLLLoss"
            weight = [0.1, 0.1]
        """
        conf = CriterionConf(**toml.loads(load))
        conf.target == "torch.nn.NLLLoss"
        assert conf.weight == [0.1, 0.1]

    @pytest.mark.skip(reason="weights values are not validated anymore")
    def test_criterion_failed_on_negative_weight(self):
        load = """
            target = "torch.nn::NLLLoss"
            weight = [-0.1, 0.1]
        """
        with pytest.raises(ValidationError):
            _ = CriterionConf(**toml.loads(load))

    def test_get_criterion_proper_parent_class(self):
        load = """
            target = "torch.nn::MSELoss"
        """
        conf = CriterionConf(**toml.loads(load))
        assert issubclass(type(conf.criterion), torch.nn.Module)

    def test_get_criterion_failed_on_wrong_parent_class(self):
        load = """
            target = "tests.dummy_module::A"
        """
        with pytest.raises(
            ValidationError,
            match=r"target class of the criterion must be a subclass of*",
        ):
            _ = CriterionConf(**toml.loads(load))


class TestDatasetConf:
    def test_get_dataset_fail_on_wrong_parent_class(self):
        load = """
            target = "tests.dummy_module::DummyDatasetModuleWrong"
        """
        with pytest.raises(
            ValidationError,
            match=r"target class of the dataset module must be a subclass of*",
        ):
            _ = DatasetConf(**toml.loads(load))

    def test_define_trainval_and_train_loader(self):
        load = """
            target = "tests.dummy_module::DummyDatasetModule"

            [trainval]
            root_dir = "..."

            [train.loader]
            batch_size = 100
        """
        conf = DatasetConf(**toml.loads(load))
        assert conf.trainval is not None
        assert conf.trainval.arguments == {"root_dir": "..."}
        assert conf.train is not None
        assert conf.train.loader == {"batch_size": 100}

    def test_define_val_train_test_predict_none_on_trainval_defined(self):
        load = """
            target = "tests.dummy_module::DummyDatasetModule"

            [trainval]
            root_dir = "..."
        """
        conf = DatasetConf(**toml.loads(load))
        assert conf.trainval is not None
        assert conf.train is None
        assert conf.validation is None
        assert conf.test is None
        assert conf.predict is None


class TestConf:
    @pytest.fixture
    def full_conf_dict(self, base_conf_txt):
        entries = base_conf_txt + """
        [metrics]
        Precision = {target = "torchmetrics::Precision", task = "multiclass", num_classes=10}
        """
        yield toml.loads(entries)

    def test_conf_parse(self, base_conf_txt):
        load = base_conf_txt + """
        [metrics]
        Precision = {target = "torchmetrics::Precision", task = "multiclass", num_classes = 10}
        FBetaScore = {target = "torchmetrics::Recall", task = "multiclass", num_classes = 10, beta = 0.1}        
        """
        _ = Conf(**toml.loads(load))

    @pytest.mark.skipif(
        sys.version_info < (3, 11), reason="test for Python < 3.11"
    )
    def test_fail_on_duplicated_key_name(self, base_conf_txt):
        load = base_conf_txt + """
        [metrics]
        Precision = {target = "torchmetrics::Precision", task = "multiclass", num_classes = 10}
        Precision = {target = "torchmetrics::Precision", task = "multiclass", num_classes = 10}
        """
        with pytest.raises(ValueError, match="Duplicate keys!"):
            _ = Conf(**toml.loads(load))

    @pytest.mark.skipif(
        sys.version_info > (3, 10), reason="test for Python > 3.10"
    )
    def test_fail_on_duplicated_key_name(self, base_conf_txt):
        load = base_conf_txt + """
        [metrics]
        Precision = {target = "torchmetrics::Precision", task = "multiclass", num_classes=10}
        Precision = {target = "torchmetrics::Precision", task = "multiclass", num_classes=10}
        """
        with pytest.raises(ValueError, match="Cannot overwrite a value.*"):
            _ = Conf(**toml.loads(load))

    @pytest.mark.skip(
        reason=(
            "dots are not supported in TOML keys. cannot pass fully qualified"
            " name"
        )
    )
    def test_conf_custom_metric(self, base_conf_txt):
        load = base_conf_txt + """
        [metrics]
        test.dummy_module.CustomMetric = {}
        """
        conf = Conf(**toml.loads(load))

    @pytest.mark.skip(
        reason=(
            "dots are not supported in TOML keys. cannot pass fully qualified"
            " name"
        )
    )
    def test_conf_custom_metric_fail_on_wrong_parentclass(self, base_conf_txt):
        load = base_conf_txt + """
        [metrics]
        teset.dummy_module.CustomMetricWrong = {}
        """
        with pytest.raises(ValidationError, match="duplicate"):
            _ = Conf(**toml.loads(load))

    def test_conf_fail_on_nonexisting_metric(self, base_conf_txt):
        load = base_conf_txt + """
        [metrics]
        MyMetric = {target = "torchmetrics::NonExistingMetric"}
        """
        with pytest.raises(
            AttributeError,
            match=(
                r"module 'torchmetrics' has no attribute 'NonExistingMetric'"
            ),
        ):
            _ = Conf(**toml.loads(load))

    def test_conf_fail_on_monitoring_undefined_metric(self, base_conf_txt):
        load = base_conf_txt + """
        [metrics]
        Precision = {target = "torchmetrics::Precision"}
        """
        load_dict = toml.loads(load)
        load_dict["training"]["checkpoint"]["monitor"] = {
            "metric": "Recall",
            "stage": "val",
        }
        with pytest.raises(ValidationError):
            _ = Conf(**load_dict)

    def test_conf_get_metric_obj_failed_on_missing_target(self, base_conf_txt):
        load = base_conf_txt + """
        [metrics]
        Precision = {}
        """
        with pytest.raises(
            ValidationError,
            match=r".*`target` is not defined for the metric `Precision`.*",
        ):
            conf = Conf(**toml.loads(load))

    def test_conf_get_metric_obj_failed_on_missing_task(self, base_conf_txt):
        load = base_conf_txt + """
        [metrics]
        Precision = {target = "torchmetrics::Precision"}
        """
        conf = Conf(**toml.loads(load))
        with pytest.raises(
            TypeError,
            match=(
                r"Precision.\_\_new\_\_\(\) missing 1 required positional"
                r" argument*"
            ),
        ):
            conf.metrics_obj

    def test_conf_get_metric_obj(self, base_conf_txt):
        load = base_conf_txt + """
        [metrics]
        Precision = {target = "torchmetrics::Precision", task = "multiclass", num_classes = 10}
        """
        conf = Conf(**toml.loads(load))
        metrics = conf.metrics_obj
        assert "precision" in metrics
        assert isinstance(metrics["precision"], tm.Metric)

    def test_conf_schedulers_single_preconfigured_schedulers_classes(
        self, full_conf_dict, dummy_optimizer
    ):
        conf = Conf(**full_conf_dict)
        assert isinstance(conf.training.preconfigured_schedulers_classes, list)

    def test_conf_schedulers_preconfigured_schedulers_classes(
        self, full_conf_dict, dummy_optimizer
    ):
        conf = Conf(**full_conf_dict)
        scheduler = conf.training.preconfigured_schedulers_classes[0](
            dummy_optimizer
        )
        assert isinstance(scheduler, torch.optim.lr_scheduler.LRScheduler)

    def test_conf_schedulers_double_preconfigured_schedulers_classes(
        self, full_conf_dict, dummy_optimizer
    ):
        conf = Conf(**full_conf_dict)
        assert len(conf.training.preconfigured_schedulers_classes) == 1
        assert isinstance(
            conf.training.preconfigured_schedulers_classes[0](dummy_optimizer),
            torch.optim.lr_scheduler.CosineAnnealingLR,
        )

        full_conf_dict["training"]["epoch_schedulers"].append(
            {
                "target": "torch.optim.lr_scheduler::MultiStepLR",
                "milestones": [30, 80],
                "gamma": 0.1,
            }
        )
        conf = Conf(**full_conf_dict)
        assert len(conf.training.preconfigured_schedulers_classes) == 2
        assert isinstance(
            conf.training.preconfigured_schedulers_classes[0](dummy_optimizer),
            torch.optim.lr_scheduler.CosineAnnealingLR,
        )
        assert isinstance(
            conf.training.preconfigured_schedulers_classes[1](dummy_optimizer),
            torch.optim.lr_scheduler.MultiStepLR,
        )

    def test_use_base_exp_name_for_metric_logging(self, base_conf_txt_full):
        load = base_conf_txt_full + """
        [logging]
        type = "csv"
        """
        conf = Conf(**toml.loads(load))
        assert "name" in conf.logging.arguments
        assert (
            conf.logging.arguments["name"]
            == "handwritten_digit_classification"
        )

    def test_dont_override_exp_name_with_base_if_provided(
        self, base_conf_txt_full
    ):
        load = base_conf_txt_full + """
        [logging]
        type = "csv"
        name = "logging_exp_name"
        """
        conf = Conf(**toml.loads(load))
        assert "name" in conf.logging.arguments
        assert conf.logging.arguments["name"] == "logging_exp_name"


class TestLogging:
    @pytest.mark.parametrize(
        "lvl", ["debug", "warn", "info", "critical", "error"]
    )
    def test_get_log_lowercase(self, lvl):
        load = f"""
    level = "{lvl}"
        """
        assert LoggingConf(**toml.loads(load)).level == lvl.upper()

    def test_get_log_fail_on_wrong_lvl(self):
        load = f"""
    level = "not-existing"
        """
        with pytest.raises(ValidationError, match=r""):
            _ = LoggingConf(**toml.loads(load)).level

    @pytest.mark.parametrize(
        "metric_logger_nick",
        ["comet", "csv", "mlflow", "neptune", "tensorboard", "wandb"],
    )
    def testmetric_logger_type_available(self, metric_logger_nick):
        load = f"""
            type = "{metric_logger_nick}"
        """
        LoggingConf(**toml.loads(load))

    def test_fail_on_wrongmetric_logger_type(self):
        load = """
            type = "not_supported"
        """
        with pytest.raises(
            ValidationError, match=".*unexpected value; permitted:.*"
        ):
            LoggingConf(**toml.loads(load))

    def test_provide_kwargs_for_metric_logger(self):
        load = """
            type = "csv"
            arg1 = 1
            arg2 = "2"
        """
        conf = LoggingConf(**toml.loads(load))
        assert conf.arguments.get("arg1") == 1
        assert conf.arguments.get("arg2") == "2"

    @pytest.mark.parametrize(
        "log_nick, log_class",
        [
            ("comet", pl_logs.CometLogger),
            ("csv", pl_logs.CSVLogger),
            ("mlflow", pl_logs.MLFlowLogger),
            ("neptune", pl_logs.NeptuneLogger),
            ("tensorboard", pl_logs.TensorBoardLogger),
            ("wandb", pl_logs.WandbLogger),
        ],
    )
    def test_metric_logger_class(self, log_nick, log_class):
        load = f"""
            type = "{log_nick}"
        """
        conf = LoggingConf(**toml.loads(load))
        assert conf.metric_logger_type == log_class

    def test_default_on_empty_string(self):
        load = ""
        conf = LoggingConf(**toml.loads(load))
        assert conf.level
        assert conf.format_
        assert conf.type_

    @pytest.mark.parametrize(
        "log_nick, attr_name",
        [
            ("comet", "experiment_name"),
            ("csv", "name"),
            ("mlflow", "experiment_name"),
            ("neptune", "name"),
            ("tensorboard", "name"),
            ("wandb", "name"),
        ],
    )
    def test_update_project_name_if_undefined(self, log_nick, attr_name):
        EXP_NAME = "new_exp_name"
        load = f"""
            type = "{log_nick}"
        """
        conf = LoggingConf(**toml.loads(load))
        conf.maybe_update_experiment_name(EXP_NAME)
        assert conf.arguments[attr_name] == EXP_NAME

    @pytest.mark.parametrize(
        "log_nick, attr_name",
        [
            ("comet", "experiment_name"),
            ("csv", "name"),
            ("mlflow", "experiment_name"),
            ("neptune", "name"),
            ("tensorboard", "name"),
            ("wandb", "name"),
        ],
    )
    def test_does_not_override_project_name_if_defined(
        self, log_nick, attr_name
    ):
        EXP_NAME = "another_new_exp"
        OVERRIDE_EXP_NAME = "new_exp_name"
        load = f"""
            type = "{log_nick}"
            {attr_name} = "{EXP_NAME}"
        """
        conf = LoggingConf(**toml.loads(load))
        conf.maybe_update_experiment_name(OVERRIDE_EXP_NAME)
        assert conf.arguments[attr_name] == EXP_NAME

    def test_properly_parse_attributes(self):
        load = """
        type = "csv"
        level = "info"
        format = "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
        """
        conf = LoggingConf(**toml.loads(load))
        assert conf.type_ == "csv"
        assert conf.level == "INFO"
        assert (
            conf.format_
            == "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
        )



from torch.utils.data import random_split
from torchvision.datasets import MNIST

from kit4dl import Kit4DLAbstractDataModule


class MNISTCustomDataset(Kit4DLAbstractDataModule):
    def prepare_trainvaldataset(self, root_dir: str):
        dset = MNIST(root=root_dir, train=True, download=True)
        train_dset, val_dset = random_split(dset, [0.8, 0.2])
        return (train_dset, val_dset)

    def prepare_testdataset(self, root_dir: str):
        return MNIST(root=root_dir, train=False, download=True)



"""A module with configuration classes."""
import os
import warnings
from abc import ABC
from functools import partial
from inspect import signature
from typing import Any, Callable, Literal

import lightning.pytorch.loggers as pl_logs
import torch
import torchmetrics as tm
from pydantic import BaseModel, Field, root_validator, validator

import kit4dl.io as io_
from kit4dl import utils as ut
from kit4dl.kit4dl_types import FullyQualifiedName
from kit4dl.nn.validators import (
    validate_class_exists,
    validate_cuda_device_exists,
)


# ################################
#           ABSTRACT
# ################################
class _AbstractClassWithArgumentsConf(
    ABC, BaseModel, extra="allow", allow_mutation=False
):
    target: FullyQualifiedName | str
    arguments: dict[str, Any]

    _validate_class_exists = validator("target", allow_reuse=True)(
        validate_class_exists
    )

    @root_validator(pre=True)
    def _build_model_arguments(cls, values: dict[str, Any]) -> dict[str, Any]:
        if "arguments" in values:
            return values
        arguments = {
            k: v for k, v in values.items() if k not in cls.__fields__
        }
        values = {k: v for k, v in values.items() if k in cls.__fields__}
        values["arguments"] = arguments
        return values


# ################################
#  Basic experiment configuration
# ################################
class BaseConf(BaseModel):
    """Base configuration model for the experiment."""

    seed: int = Field(default=0, ge=0)
    cuda_id: int | None = None
    experiment_name: str

    _assert_cuda_device = validator("cuda_id", allow_reuse=True)(
        validate_cuda_device_exists
    )

    @property
    def device(self) -> torch.device:
        """Get PyTorch device based on the provided configuration."""
        if self.cuda_id is None:
            return torch.device("cpu")
        return torch.device(f"cuda:{self.cuda_id}")

    @property
    def accelerator_device_and_id(self) -> tuple[str, list[int] | str]:
        """Get an accelerator name and its ID.

        Get accelerator name: cpu or gpu and ID of the device.
        For CPU returns instead of ID "auto" is returned.
        """
        if self.cuda_id is None:
            return ("cpu", "auto")
        return ("gpu", [self.cuda_id])


# ################################
#  Neural network model configuration
# ################################
class ModelConf(_AbstractClassWithArgumentsConf):
    """Model configuration class."""

    @validator("target")
    def _check_if_target_has_expected_parent_class(cls, value):
        from kit4dl.nn.base import (  # pylint: disable=import-outside-toplevel
            Kit4DLAbstractModule,
        )

        target_class = io_.import_and_get_attr_from_fully_qualified_name(value)
        assert issubclass(target_class, Kit4DLAbstractModule), (
            f"target class must be a subclass of `{Kit4DLAbstractModule}`"
            " class!"
        )
        return value

    @property
    def model_class(self) -> type:
        """Get Python class of the model defined in the configuration."""
        target_class = io_.import_and_get_attr_from_fully_qualified_name(
            self.target
        )
        return target_class


# ################################
#       Optimizer configuration
# ################################
class OptimizerConf(_AbstractClassWithArgumentsConf):
    """Optimizer configuration class."""

    lr: float = Field(gt=0)

    @validator("target")
    def _check_if_target_has_expected_parent_class(cls, value):
        target_class = io_.import_and_get_attr_from_fully_qualified_name(value)
        assert issubclass(target_class, torch.optim.Optimizer), (
            "target class of the optimizer must be a subclass of"
            f" `{torch.optim.Optimizer}` class!"
        )
        return value

    @property
    def optimizer(self) -> Callable[..., torch.optim.Optimizer]:
        """Get `partial` object of the preconfigured optimizer."""
        target_class = io_.import_and_get_attr_from_fully_qualified_name(
            self.target
        )
        return partial(target_class, lr=self.lr, **self.arguments)


# ################################
#       Checkpoint configuration
# ################################
class CheckpointConf(BaseModel):
    """Checkpoint configuration class."""

    path: str
    monitor: dict[str, str]
    filename: str
    mode: Literal["min", "max"] = "max"
    save_top_k: int = Field(1, ge=1)
    save_weights_only: bool = True
    every_n_epochs: int = Field(1, ge=1)
    save_on_train_epoch_end: bool = False

    @validator("path")
    def _warn_on_existing_path(cls, path):
        if os.path.exists(path) and len(os.listdir(path)) > 0:
            warnings.warn(
                f"directory {path} exists and is not empty", ResourceWarning
            )
        return path

    @validator("monitor")
    def _assert_required_monitor_keys_defined(cls, monitor):
        assert "metric" in monitor, "`metric` key is missing"
        assert (
            "stage" in monitor
        ), "`stage` key is missing. define `stage='train'` or `stage='val'`"
        assert monitor["stage"] in {"train", "val"}
        return monitor

    @property
    def monitor_metric_name(self) -> str:
        """Returns the name of the metric to be monitored."""
        return "_".join(
            [self.monitor["stage"].lower(), self.monitor["metric"].lower()]
        )


# ################################
#     Criterion configuration
# ################################
class CriterionConf(BaseModel):
    """Criterion configuration class."""

    target: FullyQualifiedName
    weight: list[float] | None = None

    _validate_class_exists = validator("target", allow_reuse=True)(
        validate_class_exists
    )

    @validator("target")
    def _check_if_target_has_expected_parent_class(cls, value):
        target_class = io_.import_and_get_attr_from_fully_qualified_name(value)
        assert issubclass(target_class, torch.nn.Module), (
            "target class of the criterion must be a subclass of"
            f" `{torch.nn.Module}` class!"
        )
        return value

    @root_validator(skip_on_failure=True)
    def _match_weight(cls, values):
        if values.get("weight"):
            criterion_class = (
                io_.import_and_get_attr_from_fully_qualified_name(
                    values["target"]
                )
            )
            assert "weight" in signature(criterion_class).parameters, (
                "`weight` parameter is not defined for the criterion"
                f" `{criterion_class}`"
            )
        return values

    @property
    def criterion(self) -> torch.nn.Module:
        """Get the torch.nn.Module with the criterion function."""
        target_class = io_.import_and_get_attr_from_fully_qualified_name(
            self.target
        )
        if self.weight is not None:
            weight_tensor = torch.FloatTensor(self.weight)
            return target_class(weight=weight_tensor)
        return target_class()


# ################################
#     Training configuration
# ################################
class TrainingConf(BaseModel):
    """Training procedure configuration class."""

    epochs: int = Field(gt=0)
    epoch_schedulers: list[dict[str, Any]] = Field(default_factory=list)
    checkpoint: CheckpointConf | None = None
    optimizer: OptimizerConf
    criterion: CriterionConf
    arguments: dict[str, Any]

    @root_validator(pre=True)
    def _build_model_arguments(cls, values: dict[str, Any]) -> dict[str, Any]:
        if "arguments" in values:
            return values
        arguments = {
            k: v for k, v in values.items() if k not in cls.__fields__
        }
        values = {k: v for k, v in values.items() if k in cls.__fields__}
        values["arguments"] = arguments
        return values

    @validator("epoch_schedulers", each_item=True)
    def _assert_epoch_scheduler_exist(cls, sch):
        assert "target" in sch
        validate_class_exists(sch["target"])
        return sch

    @property
    def preconfigured_schedulers_classes(
        self,
    ) -> list[Callable]:
        """Get a list of preconfigured schedulers."""
        schedulers: list[Callable] = []

        for sch in self.epoch_schedulers:
            sch_copy = sch.copy()
            schedulers.append(
                partial(
                    io_.import_and_get_attr_from_fully_qualified_name(
                        sch_copy.pop("target")
                    ),
                    **sch_copy,
                )
            )
        return schedulers


# ################################
#     Validation configuration
# ################################
class ValidationConf(BaseModel):
    """Validation procedure configuration class."""

    run_every_epoch: int = Field(gt=0)


# ################################
#     Split dataset configuration
# ################################
class SplitDatasetConf(BaseModel):
    """Configuration class with dataset split arguments."""

    loader: dict[str, Any] = Field(default_factory=dict)
    arguments: dict[str, Any]

    @root_validator(pre=True)
    def _build_model_arguments(cls, values: dict[str, Any]) -> dict[str, Any]:
        if "arguments" in values:
            return values
        arguments = {
            k: v for k, v in values.items() if k not in cls.__fields__
        }
        values = {k: v for k, v in values.items() if k in cls.__fields__}
        values["arguments"] = arguments
        return values


# ################################
#     Dataset configuration
# ################################
class DatasetConf(BaseModel):
    """Dataset configuration class."""

    target: FullyQualifiedName | str
    train: SplitDatasetConf | None = None
    validation: SplitDatasetConf | None = None
    trainval: SplitDatasetConf | None = None
    test: SplitDatasetConf | None = None
    predict: SplitDatasetConf | None = None
    arguments: dict[str, Any]

    @root_validator(pre=True)
    def _build_model_arguments(cls, values: dict[str, Any]) -> dict[str, Any]:
        if "arguments" in values:
            return values
        arguments = {
            k: v for k, v in values.items() if k not in cls.__fields__
        }
        values = {k: v for k, v in values.items() if k in cls.__fields__}
        values["arguments"] = arguments
        return values

    @validator("target")
    def _check_if_target_has_expected_parent_class(cls, value):
        from kit4dl.dataset import (  # pylint: disable=import-outside-toplevel
            Kit4DLAbstractDataModule,
        )

        target_class = io_.import_and_get_attr_from_fully_qualified_name(value)
        assert issubclass(target_class, Kit4DLAbstractDataModule), (
            "target class of the dataset module must be a subclass of"
            f" `{Kit4DLAbstractDataModule}` class!"
        )
        return value

    @property
    def datamodule_class(self) -> type:
        """Get Python class of the data module."""
        target_class = io_.import_and_get_attr_from_fully_qualified_name(
            self.target
        )
        return target_class


# ################################
#     Logging configuration
# ################################
_LOGGERS_NICKNAMES: dict[str, str] = {
    "comet": "CometLogger",
    "csv": "CSVLogger",
    "mlflow": "MLFlowLogger",
    "neptune": "NeptuneLogger",
    "tensorboard": "TensorBoardLogger",
    "wandb": "WandbLogger",
}


class LoggingConf(
    BaseModel, allow_population_by_field_name=True, extra="allow"
):
    """Logging configuration class."""

    type_: Literal[
        "comet", "csv", "mlflow", "neptune", "tensorboard", "wandb"
    ] = Field("csv", alias="type")
    level: Literal["DEBUG", "INFO", "WARN", "ERROR", "CRITICAL"] | None = (
        "INFO"
    )
    format_: str | None = Field("%(asctime)s - %(message)s", alias="format")
    arguments: dict[str, Any] = Field(default_factory=dict)

    @root_validator(pre=False)
    def _build_model_arguments(cls, values: dict[str, Any]) -> dict[str, Any]:
        arguments = {
            k: v for k, v in values.items() if k not in cls.__fields__
        }
        values = {k: v for k, v in values.items() if k in cls.__fields__}
        values["arguments"] = arguments
        return values

    @validator("level", pre=True)
    def _match_log_level(cls, item):
        return item.upper()

    def maybe_update_experiment_name(self, experiment_name: str) -> None:
        """Update experiment name for the chosen metric logger."""
        ltype = self.metric_logger_type
        if issubclass(ltype, (pl_logs.CometLogger, pl_logs.MLFlowLogger)):
            self.arguments.setdefault("experiment_name", experiment_name)
        elif issubclass(
            ltype,
            (
                pl_logs.CSVLogger,
                pl_logs.NeptuneLogger,
                pl_logs.TensorBoardLogger,
                pl_logs.WandbLogger,
            ),
        ):
            self.arguments.setdefault("name", experiment_name)
        else:
            raise TypeError(
                f"logger of type `{self.metric_logger_type}` is undefined!"
            )
        if issubclass(ltype, pl_logs.CSVLogger):
            self.arguments.setdefault("save_dir", "./csv_logs")

    @property
    def metric_logger_type(self) -> type:
        """Get type of the selected logger."""
        return getattr(pl_logs, _LOGGERS_NICKNAMES[self.type_])


# ################################
#     Complete configuration
# ################################
class Conf(BaseModel):
    """Conf class being the reflection of the configuration TOML file."""

    base: BaseConf
    logging: LoggingConf = Field(default_factory=LoggingConf)  # type: ignore[arg-type]
    model: ModelConf
    metrics: dict[str, dict[str, Any]] | None = Field(default_factory=dict)
    training: TrainingConf
    validation: ValidationConf
    dataset: DatasetConf

    def __init__(self, root_dir: str | None = None, **kwargs):
        if root_dir:
            kwargs = Conf.override_with_abs_target(root_dir, kwargs)
        super().__init__(**kwargs)

    @validator("logging")
    def _update_experiment_name_if_undefined(cls, value: LoggingConf, values):
        if "base" not in values:
            return value
        value.maybe_update_experiment_name(values["base"].experiment_name)
        return value

    @validator("metrics")
    def _validate_metrics_class_exist(cls, values):
        if not values:
            return None
        for metric_name, metric_dict in values.items():
            assert (
                "target" in metric_dict
            ), f"`target` is not defined for the metric `{metric_name}`"
            target_class = io_.import_and_get_attr_from_fully_qualified_name(
                metric_dict["target"]
            )
            # TODO: issubclass for torchmetrics metric does not work
            # as __bases__ for metrics in `object`. Method issubclass
            # can be used for custom metrics
            _, attr_name = io_.split_target(metric_dict["target"])
            assert issubclass(target_class, tm.Metric) or hasattr(
                tm, attr_name
            ), (
                "custom metrics need to be subclasses of `torchmetrics.Metric`"
                " class!"
            )
        return values

    @root_validator(skip_on_failure=True)
    def _check_metric_in_checkpoint_is_defined(cls, values):
        monitored_metric_name = values["training"].checkpoint.monitor["metric"]
        assert monitored_metric_name in values["metrics"].keys(), (
            f"metric `{monitored_metric_name}` is not defined. did you forget"
            f" to define `{monitored_metric_name}` in [metrics]?"
        )
        return values

    @property
    def metrics_obj(self) -> dict[str, tm.Metric]:
        """Get the dictionary of the metric name and torchmetrics.Metric."""
        if not self.metrics:
            raise ValueError("metrics are not defined!")
        metric_obj: dict = {}
        for metric_name, metric_args in self.metrics.items():
            metric_args_copy = metric_args.copy()
            metric_obj[metric_name.lower()] = (
                io_.import_and_get_attr_from_fully_qualified_name(
                    metric_args_copy.pop("target")
                )(**metric_args_copy)
            )
        return metric_obj

    @classmethod
    def override_with_abs_target(cls, root_dir: str, entries: dict) -> dict:
        """Replace in-place `target` valuesof the `entries`.

        Parameters
        ----------
        root_dir : str
            Root dir to consider as an initial part of the absolute path
        entries : dict
            Dictionary of values to replace `target` key

        Returns
        -------
        replaced : dict
            In-place modified `entries` dictionary
        """
        replace_logic = partial(io_.maybe_get_abs_target, root_dir=root_dir)
        entries = ut.replace_item_recursively(
            entries, key="target", replace=replace_logic
        )
        return entries



"""A module with the base class of modules supported by Kit4DL."""
import logging
from abc import ABC, abstractmethod
from typing import Any

import lightning.pytorch as pl
import torch

from kit4dl.metric import MetricStore
from kit4dl.mixins import LoggerMixin
from kit4dl.nn.confmodels import Conf
from kit4dl.stages import Stage


class Kit4DLAbstractModule(
    ABC, pl.LightningModule, LoggerMixin
):  # pylint: disable=too-many-ancestors
    """Base abstract class for Kit4DL modules."""

    def __init__(self, *, conf: Conf) -> None:
        super().__init__()
        assert conf, "`conf` argument cannot be `None`"
        self._criterion: torch.nn.Module | None = None
        self._conf: Conf = conf

        self._logger = logging.getLogger("lightning")
        self.configure_logger()

        self._configure_metrics()
        self._configure_criterion()
        self.configure(**self._conf.model.arguments)
        self.save_hyperparameters(self._conf.dict())

    def configure_logger(self) -> None:
        """Configure logger based on the configuration passed to the class.

        The methods configure the logger format and sets it to all
        the handlers.
        """
        self._logger.setLevel(self._conf.logging.level)  # type: ignore[arg-type]
        if self._conf.logging.format_:
            formatter = logging.Formatter(self._conf.logging.format_)
            for handler in self._logger.handlers:
                handler.setFormatter(formatter)
        for handler in self._logger.handlers:
            handler.setLevel(self._conf.logging.level)  # type: ignore[arg-type]

    @property
    def _kit4dl_logger(self) -> logging.Logger:
        return self._logger

    @abstractmethod
    def configure(self, *args: Any, **kwargs: Any) -> None:
        """Configure the architecture of the neural network.

        Parameters
        ----------
        *args: Any
            List of positional arguments to setup the network architecture
        **kwargs : Any
            List of named arguments required to setup the network architecture

        Examples
        --------
        ```python
        def configure(self, input_dims, output_dims) -> None:
            self.fc1 = nn.Sequential(
                nn.Linear(input_dims, output_dims),
            )
        ```
        """
        raise NotImplementedError

    @abstractmethod
    def run_step(self, batch, batch_idx) -> tuple[torch.Tensor, ...]:
        """Carry out single train/validation/test step for the given `batch`.

        Return a tuple of two `torch.Tensor`'s: true labels and predicted scores.
        If you need to define separate logic for validation or test step,
        implement `val_step` or `test_step` methods, respectivelly.

        Parameters
        ----------
        batch : torch.Tensor or tuple of torch.Tensor or list of torch.Tensor
            The output of the Dataloader
        batch_idx : int
            Index of the batch

        Returns
        -------
        result : tuple of `torch.Tensor`
            A tuple of 2 or 3 items:
            - if a tuple of 2 elements:
                1. `torch.Tensor` of ground-truth labels,
                2. `torch.Tensor` output of the network,
            - if a tuple of 3 elements:
                1. `torch.Tensor` of ground-truth labels,
                2. `torch.Tensor` output of the network,
                4. `torch.Tensor` with loss value.

        Examples
        --------
        ```python
        ...
        def run_step(self, batch, batch_idx) -> tuple[torch.Tensor,  ...]:
            feature_input, label_input = batch
            scores = self(feature_input)
            return (label_input, scores)
        ```

        ```python
        ...
        def run_step(self, batch, batch_idx) -> tuple[torch.Tensor, ...]:
            feature_input, label_input = batch
            scores = self(feature_input)
            loss = super().compute_loss(prediction=logits, target=is_fire)
            return (label_input, scores, loss)
        ```
        """
        raise NotImplementedError

    def run_val_step(self, batch, batch_idx) -> tuple[torch.Tensor, ...]:
        """Carry out single validation step for the given `batch`.

        Return a tuple of two `torch.Tensor`'s: true labels and predicted scores.
        If you need to define separate logic for validation or test step,
        implement `val_step` or `test_step` methods, respectivelly.

        Parameters
        ----------
        batch : torch.Tensor or tuple of torch.Tensor or list of torch.Tensor
            The output of the Dataloader
        batch_idx : int
            Index of the batch

        Returns
        -------
        result : tuple of `torch.Tensor`
            A tuple of 2 or 3 items:
            - if a tuple of 2 elements:
                1. `torch.Tensor` of ground-truth labels,
                2. `torch.Tensor` output of the network,
            - if a tuple of 3 elements:
                1. `torch.Tensor` of ground-truth labels,
                2. `torch.Tensor` output of the network,
                4. `torch.Tensor` with loss value.

        Examples
        --------
        ```python
        ...
        def run_val_step(self, batch, batch_idx) -> tuple[torch.Tensor,  ...]:
            feature_input, label_input = batch
            scores = self(feature_input)
            return (label_input, scores)
        ```

        ```python
        ...
        def run_val_step(self, batch, batch_idx) -> tuple[torch.Tensor, ...]:
            feature_input, label_input = batch
            scores = self(feature_input)
            loss = super().compute_loss(prediction=logits, target=is_fire)
            return (label_input, scores, loss)
        ```
        """
        return self.run_step(batch, batch_idx)

    def run_test_step(self, batch, batch_idx) -> tuple[torch.Tensor, ...]:
        """Carry out single test step for the given `batch`.

        Return a tuple of two `torch.Tensor`'s: true labels and predicted scores.
        If you need to define separate logic for validation or test step,
        implement `val_step` or `test_step` methods, respectivelly.

        Parameters
        ----------
        batch : torch.Tensor or tuple of torch.Tensor or list of torch.Tensor
            The output of the Dataloader
        batch_idx : int
            Index of the batch

        Returns
        -------
        result : tuple of `torch.Tensor`
            A tuple of 2 or 3items:
            - if a tuple of 2 elements:
                1. `torch.Tensor` of ground-truth labels,
                2. `torch.Tensor` output of the network,
            - if a tuple of 3 elements:
                1. `torch.Tensor` of ground-truth labels,
                2. `torch.Tensor` output of the network,
                4. `torch.Tensor` with loss value.

        Examples
        --------
        ```python
        ...
        def run_test_step(self, batch, batch_idx) -> tuple[torch.Tensor,  ...]:
            feature_input, label_input = batch
            scores = self(feature_input)
            return (label_input, scores)
        ```

        ```python
        ...
        def run_test_step(self, batch, batch_idx) -> tuple[torch.Tensor, ...]:
            feature_input, label_input = batch
            scores = self(feature_input)
            loss = super().compute_loss(prediction=logits, target=is_fire)
            return (label_input, scores, loss)
        ```
        """
        return self.run_step(batch, batch_idx)

    def run_predict_step(self, batch, batch_idx) -> torch.Tensor:
        """Carry out single predict step for the given `batch`.

        Return a `torch.Tensor` - the predicted scores.
        If not overriden, the implementation of `step` method is used.

        Parameters
        ----------
        batch : torch.Tensor or tuple of torch.Tensor or list of torch.Tensor
            The output of the Dataloader
        batch_idx : int
            Index of the batch

        Returns
        -------
        result : torch.Tensor
            The score being the output of of the network


        Note
        ----
        The function returns just score values as for prediction we do not have
        the ground-truth labels.

        Examples
        --------
        ```python
        ...
        def run_predict_step(self, batch, batch_idx) -> torch.Tensor:
            feature_input = batch
            return self(feature_input)
        ```
        """
        _, scores = self.run_step(batch, batch_idx)
        return scores

    def configure_optimizers(
        self,
    ) -> tuple[
        list[torch.optim.Optimizer],
        list[torch.optim.lr_scheduler.LRScheduler] | None,
    ]:
        """Configure optimizers and schedulers."""
        self.debug("configuring optimizers and lr epoch schedulers...")
        optimizer: torch.optim.Optimizer = (
            self._conf.training.optimizer.optimizer(self.parameters())
        )
        lr_schedulers: list = [
            scheduler(optimizer)
            for scheduler in self._conf.training.preconfigured_schedulers_classes
        ]
        self.info("selected optimizer is: %s", optimizer)
        self.info(
            "selected %d  lr schedulers: %s", len(lr_schedulers), lr_schedulers
        )
        return [optimizer], lr_schedulers

    def _configure_metrics(self) -> None:
        self.debug("configuring metrics...")
        self.train_metric_tracker = MetricStore(self._conf.metrics_obj)
        self.val_metric_tracker = MetricStore(self._conf.metrics_obj)
        self.test_metric_tracker = MetricStore(self._conf.metrics_obj)

    def _configure_criterion(self) -> None:
        self.debug("configuring criterion...")
        self._criterion = self._conf.training.criterion.criterion.to(
            self._conf.base.device
        )
        self.info("selected criterion is: %s", self._criterion)

    def compute_loss(
        self, prediction: torch.Tensor, target: torch.Tensor
    ) -> torch.Tensor:
        """Compute the loss based on the prediction and target."""
        assert self._criterion, "criterion is None"
        return self._criterion(prediction, target)

    def log_train_metrics(self) -> None:
        """Log train metrics."""
        for (
            metric_name,
            metric_value,
        ) in self.train_metric_tracker.results.items():
            stage_metric_name = f"{Stage.TRAIN}_{metric_name}"
            self.info(
                "epoch: %d metric: %s value: %s",
                self.current_epoch,
                stage_metric_name,
                metric_value,
            )
            self.log(
                stage_metric_name,
                metric_value,
                logger=True,
            )

    def log_val_metrics(self) -> None:
        """Log validation metrics."""
        for (
            metric_name,
            metric_value,
        ) in self.val_metric_tracker.results.items():
            stage_metric_name = f"{Stage.VALIDATION}_{metric_name}"
            self.info(
                "epoch: %d metric: %s value: %s",
                self.current_epoch,
                stage_metric_name,
                metric_value,
            )
            self.log(
                stage_metric_name,
                metric_value,
                logger=True,
            )

    def log_test_metrics(self) -> None:
        """Log test metrics."""
        for (
            metric_name,
            metric_value,
        ) in self.test_metric_tracker.results.items():
            stage_metric_name = f"{Stage.TEST}_{metric_name}"
            self.info(
                "epoch: %d metric: %s value: %s",
                self.current_epoch,
                stage_metric_name,
                metric_value,
            )
            self.log(
                stage_metric_name,
                metric_value,
                logger=True,
            )

    def update_train_metrics(
        self, true: torch.Tensor, predictions: torch.Tensor, loss: torch.Tensor
    ) -> None:
        """Update train metrics with true and prediction values."""
        self.train_metric_tracker.update(true=true, predictions=predictions)
        if loss:
            self.log(name=f"{Stage.TRAIN}_loss", value=loss, logger=True)

    def update_val_metrics(
        self, true: torch.Tensor, predictions: torch.Tensor, loss: torch.Tensor
    ) -> None:
        """Update validation metrics with true and prediction values."""
        self.val_metric_tracker.update(true=true, predictions=predictions)
        if loss:
            self.log(name=f"{Stage.VALIDATION}_loss", value=loss, logger=True)

    def update_test_metrics(
        self, true: torch.Tensor, predictions: torch.Tensor, loss: torch.Tensor
    ) -> None:
        """Update test metrics with true and prediction values."""
        self.test_metric_tracker.update(true=true, predictions=predictions)
        if loss:
            self.log(name=f"{Stage.TEST}_loss", value=loss, logger=True)

    def reset_metric_trackers(self) -> None:
        """Reset all metric trackers: train, validation, and test."""
        self.train_metric_tracker.reset()
        self.val_metric_tracker.reset()
        self.test_metric_tracker.reset()

    def training_step(
        self, batch, batch_idx
    ):  # pylint: disable=arguments-differ
        """Carry out a single training step."""
        res = self.run_step(batch, batch_idx)
        match res:
            case (y_true, y_scores):
                loss = self.compute_loss(y_scores, y_true)
            case (y_true, y_scores, loss):
                pass
            case _:
                self.error("wrong size of tuple returned by `run_step`")
                raise ValueError("wrong size of tuple returned by `run_step`")
        self.update_train_metrics(true=y_true, predictions=y_scores, loss=loss)
        return loss

    def validation_step(
        self, batch, batch_idx
    ):  # pylint: disable=arguments-differ
        """Carry out a single validation step."""
        res = self.run_val_step(batch, batch_idx)
        match res:
            case (y_true, y_scores):
                loss = self.compute_loss(y_scores, y_true)
            case (y_true, y_scores, loss):
                pass
            case _:
                self.error("wrong size of tuple returned by `run_step`")
                raise ValueError("wrong size of tuple returned by `run_step`")
        self.update_val_metrics(true=y_true, predictions=y_scores, loss=loss)
        return loss

    def test_step(self, batch, batch_idx):  # pylint: disable=arguments-differ
        """Carry out a single test step."""
        res = self.run_test_step(batch, batch_idx)
        match res:
            case (y_true, y_scores):
                loss = self.compute_loss(y_scores, y_true)
            case (y_true, y_scores, loss):
                pass
            case _:
                self.error("wrong size of tuple returned by `run_step`")
                raise ValueError("wrong size of tuple returned by `run_step`")
        self.update_test_metrics(true=y_true, predictions=y_scores, loss=loss)
        return loss



"""Module with CLI for Kit4DL."""
import importlib.resources
import logging
import os
import shutil
import sys

try:
    import tomllib as toml
except ModuleNotFoundError:
    import toml  # type: ignore[no-redef]

import typer
from typing_extensions import Annotated

from kit4dl import context
from kit4dl.formatting import escape_os_sep, substitute_symbols
from kit4dl.nn.confmodels import Conf
from kit4dl.nn.trainer import Trainer

# ##############################
#         CREATE CLI
# ##############################
_app = typer.Typer(name="Kit4DL")

# ##############################
#       CONFIGURE LOGGER
# ##############################
_CLI_LOG_LEVEL = "INFO"
_CLI_LOG_FORMAT = "%(asctime)s - %(name)s - %(levelname)s - %(message)s"


def _configure_logger(logger: logging.Logger):
    logger.setLevel(_CLI_LOG_LEVEL)
    formatter = logging.Formatter(_CLI_LOG_FORMAT)
    handler = logging.StreamHandler(sys.stdout)
    logger.addHandler(handler)
    for hdl in log.handlers:
        hdl.setFormatter(formatter)
        hdl.setLevel(_CLI_LOG_LEVEL)  # type: ignore[arg-type]


log = logging.getLogger("Kit4DL.CLI")
_configure_logger(log)


# ##############################
#         UTILS METHODS
# ##############################
def update_context_from_runtime(
    prj_dir: str | None = None,
) -> None:
    """Update context properties from the runtime variables."""
    context.PROJECT_DIR = prj_dir


def update_context_from_conf(conf: Conf | None = None) -> None:
    """Update context from the configuration file."""
    if not conf:
        return
    context.LOG_LEVEL = conf.logging.level
    context.LOG_FORMAT = conf.logging.format_


def _get_conf_from_file(conf_path: str, root_dir: str | None = None):
    assert conf_path and conf_path.endswith(
        ".toml"
    ), "`conf_path` needs to be TOML file!"
    text_load = substitute_symbols(conf_path, **context.get_dict())
    text_load_escaped = escape_os_sep(text_load)
    return Conf(root_dir=root_dir, **toml.loads(text_load_escaped))  # type: ignore[arg-type]


def _get_default_conf_path() -> str:
    return os.path.join(os.getcwd(), "conf.toml")


def _remove_redundant_items(path):
    shutil.rmtree(os.path.join(path, "__pycache__"), ignore_errors=True)


# ##############################
#      COMMANDS DEFINITIONS
# ##############################
@_app.command()
def init(
    name: Annotated[
        str, typer.Option(help="The name of your new project")
    ] = "new_kit4dl_project"
) -> None:
    """Create a new Kit4DL project.

    Parameters
    ----------
    name : str, optional
        The optional name of the project.
        If skipped, the deafult `new_kit4dl_project` will be used
    """
    log.info("Kit4DL Creating a new skeleton for the project: << %s >>", name)
    assert not os.path.exists(name), f"The project `{name}` already exists!"
    with importlib.resources.path(
        "kit4dl.cli._templates", "project"
    ) as empty_proj_path:
        shutil.copytree(empty_proj_path, name)
    _remove_redundant_items(name)


@_app.command()
def resume(
    checkpoint: Annotated[
        str, typer.Option(help="Path to the checkpoint file")
    ]
):
    """Resume learning from the checkpoint.

    Parameters
    ----------
    checkpoint : str
        Path to a checkpoint file
    """
    raise NotImplementedError


@_app.command()
def train(
    conf: Annotated[
        str, typer.Option(help="Path to the configuration TOML file")
    ] = _get_default_conf_path(),
    test: Annotated[
        bool, typer.Option(help="If perform testing using best weights")
    ] = True,
) -> None:
    """Train using the configuration file.

    Parameters
    ----------
    conf : str, optional
        Path to the configuration TOML file.
        If skipped, the program will search for the `conf.toml` file
        in the current working directory.
    test
    """
    log.info("Attempt to run training...")
    root_dir = os.path.dirname(conf)
    if not os.path.exists(conf):
        raise RuntimeError(
            f"the conf file: {conf} does not exist. ensure the default"
            " configuration file exist or specify --conf argument to a valid"
            " configuration file."
        )
    prj_dir = os.path.join(os.getcwd(), root_dir)
    sys.path.append(prj_dir)
    update_context_from_runtime(prj_dir=prj_dir)
    conf_ = _get_conf_from_file(conf, root_dir=root_dir)
    update_context_from_conf(conf=conf_)
    log.info("Running trainer \U0001f3ac")
    trainer = Trainer(conf=conf_).prepare()
    trainer.fit()
    log.info("Training finished \U00002728")
    if test:
        log.info("Running testing \U00002728")
        trainer.test()
        log.info("Testing finished \U00002728")


def run():
    """Run the CLI app."""
    _app()



"""Module with utilities functions and structures."""
import os
import random
from typing import Callable, Hashable

import numpy as np
import torch


def set_seed(seed: int) -> None:
    """Set seed for random numbers for NumPy and PyTorch.

    Parameters
    ----------
    seed : int
        Seed value
    """
    os.environ["PL_GLOBAL_SEED"] = str(seed)
    os.environ["PL_SEED_WORKERS"] = "1"
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)


def replace_item_recursively(
    entries: dict, key: Hashable, replace: Callable
) -> dict:
    """Replace in-place value for `key` recursively.

    Replaces recursively values for `key` by the result of the `replace`
    callable. If value of `key` is a dictionary, the method iterates over it.

    Parameters
    ----------
    entries : dict
        A dictionary or dict of dict whose values for the `key`
        are to be replaced
    key : hashable
        A key, whose values will be replaced
    replace : callable
        A function taking as an argument the value of dictionary associated
        with `key`

    Returns
    -------
    res : dict
        A modified dictionary itself

    Examples
    --------
    ```python
    >>> my_dict = {"a": 1, "b": {"a": -1}}
    >>> replace_item_recursively(my_dict, "a", lambda val: val*10)
    {'a': 10, 'b': {'a': -10}}
    ```
    """
    for it_key, value in entries.items():
        if isinstance(value, dict):
            replace_item_recursively(value, key, replace)
        if key == it_key:
            entries[key] = replace(entries[key])
    return entries



from unittest.mock import MagicMock

import pytest

from kit4dl.dataset import Kit4DLAbstractDataModule


class TestDatamodule:
    class TestDataModule(Kit4DLAbstractDataModule):
        pass

    def test_setting_extra_arguments(self):
        conf = MagicMock()
        conf.arguments = {"abc": 1}
        datamodule = TestDatamodule.TestDataModule(conf)
        assert hasattr(datamodule, "abc")
        assert datamodule.abc == 1



"""Module with mixing classes definitions."""
import logging


class LoggerMixin:
    """
    Mixing class for direct logging.

    The class provides methods for direct logging instead of
    calling `_logger` methods.`
    The methods have the same signatres as those defined in the
    `logging.Logger` class.

    Note
    ----
    The class using `LoggerMixin` need to have `_logger` defined.

    Examples
    --------
    ```
    >>> class MyClass(LoggerMixin): pass
    >>> MyClass().debug("some debug message")
    """

    _logger: logging.Logger

    def debug(self, *args, **kwargs):
        """Log on debug level."""
        self._logger.debug(*args, **kwargs)

    def info(self, *args, **kwargs):
        """Log on info level."""
        self._logger.info(*args, **kwargs)

    def warn(self, *args, **kwargs):
        """Log on warning level."""
        self._logger.warn(*args, **kwargs)

    def error(self, *args, **kwargs):
        """Log on error level."""
        self._logger.error(*args, **kwargs)

    def critical(self, *args, **kwargs):
        """Log on critical level."""
        self._logger.critical(*args, **kwargs)



from unittest.mock import patch

import pytest

from kit4dl import context


class TestSession:
    @pytest.fixture(autouse=True, scope="class")
    def set_session_attributes(self):
        context.LOG_LEVEL = "INFO"
        context.LOG_FORMAT = "%(asctime)s"
        context.PROJECT_DIR = "/work/my_dir"
        yield

    @pytest.mark.parametrize(
        "session_attr", ["LOG_LEVEL", "LOG_FORMAT", "PROJECT_DIR"]
    )
    def test_fail_on_successive_session_attribute_set(self, session_attr):
        with pytest.raises(
            RuntimeError, match=r"Session properties can be set only once"
        ):
            setattr(context, session_attr, "new_value")

    def test_fail_on_creating_new_session_attribute(self):
        with pytest.raises(
            RuntimeError, match=r"Cannot set new session property!"
        ):
            context.NEW_ATTR = 10

    def test_attribute_availability_when_setting_by_other_thread(self):
        # TODO
        pass



"""A module with metric logic."""
import numpy as np
import torchmetrics as tm


class MetricStore:
    """Class being a store for different metrics.

    It manages resetting,updating, and computing the metrics result.
    """

    __slots__ = ("_metrics",)

    def __init__(self, metrics: dict[str, tm.Metric]) -> None:
        self._metrics = metrics

    def reset(self) -> "MetricStore":
        """Reset metrics to the init state."""
        for met in self._metrics.values():
            met.reset()
        return self

    def update(self, true, predictions) -> "MetricStore":
        """Update metric with `true` and `predictions` values.

        Parameters
        ----------
        true : torch.Tensor
            Tensor with true labels
        predictions : torch.Tensor
            Tensor with predictions
        """
        for met in self._metrics.values():
            met.update(predictions, true)
        return self

    @property
    def results(self) -> dict[str, float]:
        """Get dictionary with metric values."""
        return {
            metric_name: (
                np.nan
                if metric._update_count  # pylint: disable=protected-access
                == 0
                else metric.compute().item()
            )
            for metric_name, metric in self._metrics.items()
        }



from unittest.mock import MagicMixin, patch

import pytest

from kit4dl.nn.trainer import Trainer, set_seed
from tests.fixtures import conf, true_conf


class TestTrainer:
    @patch("kit4dl.nn.trainer.set_seed")
    def test_seed_set_run(self, mock_set_seed, conf):
        Trainer(conf)
        mock_set_seed.assert_called_with(conf.base.seed)

    def test_seed_set_deterministic_rand_values(self, conf):
        import numpy as np
        import torch

        Trainer(conf)
        arr1 = np.random.rand(1000)
        ten1 = torch.rand(10, 20, 30)

        Trainer(conf)
        arr2 = np.random.rand(1000)
        ten2 = torch.rand(10, 20, 30)
        assert np.allclose(arr1, arr2)
        assert torch.allclose(ten1, ten2)

    @patch("kit4dl.nn.trainer.Trainer._configure_trainer")
    @patch("kit4dl.nn.trainer.Trainer._configure_datamodule")
    def test_configure_model(
        self, mock_conf_trainer, mock_conf_datamodule, conf
    ):
        Trainer(conf).prepare()
        conf.model.model_class.assert_called_once_with(conf=conf)

    @patch("kit4dl.nn.trainer.Trainer._configure_trainer")
    @patch("kit4dl.nn.trainer.Trainer._configure_model")
    def test_configure_datamodule(
        self, mock_conf_trainer, mock_conf_model, conf
    ):
        Trainer(conf).prepare()
        conf.dataset.datamodule_class.assert_called_once_with(
            conf=conf.dataset
        )



import os
from unittest.mock import patch

import pytest

import kit4dl.io as io_


class TestIO:
    @pytest.mark.parametrize(
        "fc_name",
        [
            "torch.optim::Adam",
            "os::PathLike",
            "kit4dl.nn.base::Kit4DLAbstractModule",
        ],
    )
    def test_import_from_fully_qualified_name(self, fc_name):
        target_class = io_.import_and_get_attr_from_fully_qualified_name(
            fc_name
        )
        assert isinstance(target_class, type)
        module, class_name = fc_name.rsplit("::")
        assert module in str(target_class)
        assert class_name in str(target_class)

    @pytest.mark.parametrize("fc_name", ["os::listdir", "numpy.linalg::eigh"])
    def test_import_from_fully_qualified_function(self, fc_name):
        func = io_.import_and_get_attr_from_fully_qualified_name(fc_name)
        assert callable(func)
        _, func = fc_name.rsplit("::")
        assert func in str(func)

    def test_import_fail_on_not_existing_file(self):
        path = r"./not_existing.py::A"
        with pytest.raises(
            AssertionError, match=r"module: ./not_existing.py does not exist"
        ):
            _ = io_.import_and_get_attr_from_fully_qualified_name(path)

    @pytest.mark.parametrize(
        "path",
        [
            "./tests/dummy_module.py::A",
            "./tests/dummy_module.py::A",
            "./tests/dummy_module2.py::T1",
        ],
    )
    def test_import_from_relative_pyfile(self, path):
        target_class = io_.import_and_get_attr_from_fully_qualified_name(path)
        assert isinstance(target_class, type)
        file, class_name = path.rsplit("::")
        assert class_name in str(target_class)
        assert os.path.exists(file)

    @pytest.mark.skipif(
        not os.path.exists(r"D:\Projekty\CMCC\kit4dl\tests\dummy_module2.py"),
        reason="file does not exists. absolute path is not correct",
    )
    def test_import_from_absolute_pyfile(self):
        path = r"D:\Projekty\CMCC\kit4dl\tests\dummy_module2.py::T1"
        target_class = io_.import_and_get_attr_from_fully_qualified_name(path)
        assert isinstance(target_class, type)
        file, class_name = path.rsplit("::")
        assert class_name in str(target_class)
        assert os.path.exists(file)

    def test_nonoverriding_imported_classes(self):
        path_1 = "./tests/dummy_module.py::T1"
        path_2 = "./tests/dummy_module2.py::T1"

        cls1 = io_.import_and_get_attr_from_fully_qualified_name(path_1)
        cls2 = io_.import_and_get_attr_from_fully_qualified_name(path_2)

        assert id(cls1) != id(cls2)
        obj1 = cls1()
        assert hasattr(obj1, "f1")
        assert not hasattr(obj1, "f2")

        obj2 = cls2()
        assert hasattr(obj2, "f2")
        assert not hasattr(obj2, "f1")



"""Kit4DL package."""
from kit4dl.dataset import Kit4DLAbstractDataModule
from kit4dl.nn.base import Kit4DLAbstractModule

# NOTE: aliases for backward compatibility with the version 2023.08b0
MLKitAbstractDataModule = Kit4DLAbstractDataModule
MLKitAbstractModule = Kit4DLAbstractModule



import os
from unittest.mock import MagicMock, Mock, PropertyMock, patch

import pytest

try:
    import tomllib as toml
except ModuleNotFoundError:
    import toml

import torch
import torchmetrics as tm

from kit4dl.nn.base import Kit4DLAbstractModule
from tests.fixtures import conf


class TestBaseMLModel:
    @pytest.fixture
    def custom_module(self, conf):
        class CustomModule(Kit4DLAbstractModule):
            configure = MagicMock()
            run_step = MagicMock()

        obj = CustomModule(conf=conf)
        yield obj

    @pytest.fixture
    def module(self):
        from torch import nn

        class Module(Kit4DLAbstractModule):
            def configure(self, input_dims, output_dims) -> None:
                self.l1 = nn.Sequential(
                    nn.Conv1d(
                        input_dims,
                        128,
                        kernel_size=3,
                        padding="same",
                        bias=False,
                    ),
                    nn.ReLU(),
                    nn.BatchNorm1d(128),
                    nn.Conv1d(
                        128,
                        output_dims,
                        kernel_size=3,
                        padding="same",
                        bias=True,
                    ),
                )

            def run_step(
                self, batch, batch_idx
            ) -> tuple[torch.Tensor, torch.Tensor]:
                x, cat, instance = batch
                res = self.l1(x.permute(0, 2, 1))
                return cat, res

        return Module

    def test_setup_metric_trackers(self, custom_module):
        assert custom_module.train_metric_tracker is not None
        assert custom_module.val_metric_tracker is not None
        assert custom_module.test_metric_tracker is not None

    def test_setup_called(self, custom_module, conf):
        custom_module.configure.assert_called_with(**conf.model.arguments)



import torchmetrics as tm

from kit4dl.dataset import Kit4DLAbstractDataModule
from kit4dl.nn.base import Kit4DLAbstractModule


class A:
    pass


class B(Kit4DLAbstractModule):
    def setup(self, input_dims, layers, dropout, output_dims):
        pass

    def run_step(self, batch, batch_idx):
        pass


class DummyDatasetModuleWrong:
    pass


class DummyDatasetModule(Kit4DLAbstractDataModule):
    def __init__(self, train_config, root_dir: str):
        super().__init__(train_config, None, None)


class CustomMetricWrong:
    pass


class CustomMetric(tm.Metric):
    pass


class T1:
    def f1():
        pass



"""Docstring of the module containing my data module."""
from typing import Any

from torch.utils.data import Dataset

from kit4dl import Kit4DLAbstractDataModule


class MyCustomDatamodule(Kit4DLAbstractDataModule):
    """My datamodule docstring."""

    def prepare_traindataset(self, *args: Any, **kwargs: Any) -> Dataset:
        """Prepare train dataset."""
        raise NotImplementedError

    def prepare_valdataset(self, *args: Any, **kwargs: Any) -> Dataset:
        """Prepare validation dataset."""
        raise NotImplementedError

    def prepare_trainvaldataset(
        self, *args: Any, **kwargs: Any
    ) -> tuple[Dataset, Dataset]:
        """Prepare train and validation dataset."""
        # NOTE: If you have single logic for train/val split,
        # implement this method and remove `prepare_traindataset`
        # and `prepare_valdataset`
        raise NotImplementedError

    def prepare_testdataset(self, *args: Any, **kwargs: Any) -> Dataset:
        """Prepare test dataset."""
        raise NotImplementedError

    def prepare_predictdataset(self, *args: Any, **kwargs: Any) -> Dataset:
        """Prepare predict dataset."""
        raise NotImplementedError



"""A module with the Kit4DL abstract dataset definition."""
import logging
from abc import ABC
from typing import Any, Callable

import lightning.pytorch as pl
from torch.utils.data import DataLoader, Dataset

from kit4dl import context
from kit4dl.mixins import LoggerMixin
from kit4dl.nn.confmodels import DatasetConf


class Kit4DLAbstractDataModule(ABC, pl.LightningDataModule, LoggerMixin):
    """The class with the logic for dataset management.

    The class provides a user with a simple interface:
    `prepare_data` is a method for downloading and preprocessing datasets
    `prepare_traindataset` is a method returning `torch.Dataset` for train
        data
    `prepare_valdataset` is a method returning `torch.Dataset` for validation
        data
    `prepare_trainvaldataset` is a method returning a tuple of two
        `torch.Dataset`s for train and validation data
        (if this method is provided, `prepare_traindataset` and
        `prepare_valdataset` shouldn't be implemented)
    `prepare_testdataset` is a method returning `torch.Dataset` for test data
    `prepare_predictdataset` is a method returning `torch.Dataset` for data
        for prediction
    """

    def __init__(self, conf: DatasetConf):
        super().__init__()
        self.conf = conf
        self.train_dataset: Dataset | None = None
        self.val_dataset: Dataset | None = None
        self.test_dataset: Dataset | None = None
        self.predict_dataset: Dataset | None = None
        self._configure_logger()
        for extra_arg_key, extra_arg_value in self.conf.arguments.items():
            self.debug(
                "setting extra user-defined argument: %s:%s",
                extra_arg_key,
                extra_arg_value,
            )
            setattr(self, extra_arg_key, extra_arg_value)

    def _configure_logger(self) -> None:
        self._logger = logging.getLogger("kit4dl.dataset")
        self._logger.setLevel(context.LOG_LEVEL)
        if context.LOG_FORMAT:
            formatter = logging.Formatter(context.LOG_FORMAT)
            for handler in self._logger.handlers:
                handler.setFormatter(formatter)
        for handler in self._logger.handlers:
            handler.setLevel(context.LOG_LEVEL)  # type: ignore[arg-type]

    def prepare_data(self):
        """Prepare dataset for train/validation/test/predict splits.

        Examples
        --------
        ```python
        class MyDatamodule(Kit4DLAbstractDataModule):

            def prepare_data(self):
                # any logic you need to perform before creating splits
                download_dataset()
        ```
        """

    def prepare_traindataset(self, *args: Any, **kwargs: Any) -> Dataset:
        """Prepare dataset for training.

        Parameters
        ----------
        *args: Any
            List of positional arguments to setup the dataset
        **kwargs : Any
            List of named arguments required to setup the dataset

        Returns
        -------
        train_datasets : Dataset
            A training dataset

        Examples
        --------
        ```python
        ...
        def prepare_traindataset(self, root_dir: str) -> Dataset:
            train_dset = MyDataset(root_dir=root_dir)
            return train_dset
        ```
        """
        raise NotImplementedError

    def prepare_valdataset(self, *args: Any, **kwargs: Any) -> Dataset:
        """Prepare dataset for validation.

        Parameters
        ----------
        *args: Any
            List of positional arguments to setup the dataset
        **kwargs : Any
            List of named arguments required to setup the dataset

        Returns
        -------
        val_datasets : Dataset
            A validation dataset

        Examples
        --------
        ```python
        ...
        def prepare_valdataset(self, root_dir: str) -> Dataset:
            val_dset = MyDataset(root_dir=root_dir)
            return val_dset
        ```
        """
        raise NotImplementedError

    def prepare_trainvaldataset(
        self, *args: Any, **kwargs: Any
    ) -> tuple[Dataset, Dataset]:
        """Prepare dataset for training and validation.

        Parameters
        ----------
        *args: Any
            List of positional arguments to setup the dataset
        **kwargs : Any
            List of named arguments required to setup the dataset

        Returns
        -------
        trainval_datasets : tuple of Dataset
            Tuple consisting of train and validation dataset

        Examples
        --------
        ```python
        ...
        def prepare_trainvaldataset(self, root_dir: str) -> tuple[Dataset, Dataset]:
            dset = MyDataset(root_dir=root_dir)
            train_dset, val_dset = random_split(dset, [500, 50])
            return train_dset, val_dset
        ```
        """
        raise NotImplementedError

    def prepare_testdataset(self, *args: Any, **kwargs: Any) -> Dataset:
        """Prepare dataset for testing.

        Parameters
        ----------
        *args: Any
            List of positional arguments to setup the dataset
        **kwargs : Any
            List of named arguments required to setup the dataset

        Returns
        -------
        test_datasets : Dataset
            A test dataset

        Examples
        --------
        ```python
        ...
        def prepare_testdataset(self, root_dir: str) -> Dataset:
            return MyDataset(root_dir=root_dir)
        ```
        """
        raise NotImplementedError

    def prepare_predictdataset(self, *args: Any, **kwargs: Any) -> Dataset:
        """Prepare dataset for predicting.

        Parameters
        ----------
        *args: Any
            List of positional arguments to setup the dataset
        **kwargs : Any
            List of named arguments required to setup the dataset

        Returns
        -------
        pred_datasets : Dataset
            A prediction dataset

        Examples
        --------
        ```python
        ...
        def prepare_predictdataset(self, root_dir: str) -> Dataset:
            return MyDataset(root_dir=root_dir)
        ```
        """
        raise NotImplementedError

    def _handle_fit_stage(self) -> None:
        if self.conf.trainval:
            assert self.conf.trainval, (
                "configuration for train-validation phases is not defined. did"
                " you forget to define `[dataset.trainval]` section in the"
                " configuration file?"
            )
            self.train_dataset, self.val_dataset = (
                self.prepare_trainvaldataset(**self.conf.trainval.arguments)
            )
        else:
            assert self.conf.train, (
                "configuration for train phase is not defined. did you forget"
                " to define `[dataset.train]` section in the configuration"
                " file?"
            )
            assert self.conf.validation, (
                "configuration for validation phase is not defined. did you"
                " forget to define `[dataset.validation]` section in the"
                " configuration file?"
            )
            self.train_dataset = self.prepare_traindataset(
                **self.conf.train.arguments
            )
            self.val_dataset = self.prepare_valdataset(
                **self.conf.validation.arguments
            )

    def _handle_test_stage(self) -> None:
        assert self.conf.test, (
            "configuration for test phase is not defined. did you forget to"
            " define `[dataset.test]` section in the configuration file?"
        )
        self.test_dataset = self.prepare_testdataset(
            **self.conf.test.arguments
        )

    def _handle_predict_stage(self) -> None:
        assert self.conf.predict, (
            "`test_config` is not defined. did you forget to define"
            " `[dataset.predict]` section in the configuration file?"
        )
        self.predict_dataset = self.prepare_predictdataset(
            **self.conf.predict.arguments
        )

    def setup(self, stage: str) -> None:
        """Set up data depending on the stage.

        The method should not be overriden unless necessary.

        Parameters
        ----------
        stage : str
            The stage of the pipeline. One out of `['fit', 'test', 'predict']`
        """
        match stage:
            case "fit":
                self._handle_fit_stage()
            case "test":
                self._handle_test_stage()
            case "predict":
                self._handle_predict_stage()

    def get_collate_fn(self) -> Callable | None:
        """Get batch collate function."""
        return None

    def get_train_collate_fn(self) -> Callable | None:
        """Get train specific collate function."""
        return self.get_collate_fn()

    def get_val_collate_fn(self) -> Callable | None:
        """Get validation specific collate function."""
        return self.get_collate_fn()

    def get_test_collate_fn(self) -> Callable | None:
        """Get test specific collate function."""
        return self.get_collate_fn()

    def get_predict_collate_fn(self) -> Callable | None:
        """Get predict specific collate function."""
        return self.get_collate_fn()

    def train_dataloader(self) -> DataLoader:
        """Prepare loader for train data."""
        assert self.conf.train, (
            "train configuration is not defined. did you forget"
            " [dataset.train] section in the configuration file?"
        )
        assert self.train_dataset is not None, (
            "did you forget to return `torch.utils.data.Dataset`instance from"
            " the `prepare_traindataset` method?"
        )
        return DataLoader(
            self.train_dataset,
            **self.conf.train.loader,
            collate_fn=self.get_train_collate_fn(),
        )

    def val_dataloader(self) -> DataLoader:
        """Prepare loader for validation data."""
        assert self.conf.validation, (
            "validation configuration is not defined. did you forget"
            " [dataset.validation] section in the configuration file?"
        )
        assert self.val_dataset is not None, (
            "did you forget to return `torch.utils.data.Dataset`instance from"
            " the `prepare_valdataset` method?"
        )
        return DataLoader(
            self.val_dataset,
            **self.conf.validation.loader,
            collate_fn=self.get_val_collate_fn(),
        )

    def test_dataloader(self) -> DataLoader:
        """Prepare loader for test data."""
        assert self.conf.test, (
            "test configuration is not defined. did you forget"
            " [dataset.test] section in the configuration file?"
        )
        assert self.test_dataset is not None, (
            "did you forget to return `torch.utils.data.Dataset`instance from"
            " the `prepare_testdataset` method?"
        )
        return DataLoader(
            self.test_dataset,
            **self.conf.test.loader,
            collate_fn=self.get_test_collate_fn(),
        )

    def predict_dataloader(self) -> DataLoader:
        """Prepare loader for prediction data."""
        assert self.conf.predict, (
            "validation configuration is not defined. did you forget"
            " [dataset.predict] section in the configuration file?"
        )
        assert self.predict_dataset is not None, (
            "did you forget to return `torch.utils.data.Dataset`instance from"
            " the `prepare_predictdataset` method?"
        )
        return DataLoader(
            self.predict_dataset,
            **self.conf.predict.loader,
            collate_fn=self.get_predict_collate_fn(),
        )

    def numpy_train_dataloader(self):
        """Prepare loader for train data for models accepting `numpy.ndarray`."""
        raise NotImplementedError

    def numpy_val_dataloader(self):
        """Prepare loader for val data for models accepting `numpy.ndarray`."""
        raise NotImplementedError

    def numpy_testdataloader(self):
        """Prepare loader for test data for models accepting `numpy.ndarray`."""
        raise NotImplementedError

    def numpy_predictdataloader(self):
        """Prepare loader for pred data for models accepting `numpy.ndarray`."""
        raise NotImplementedError



import warnings
from contextlib import suppress
from functools import partial

import numpy as np
import pytest
import sklearn.metrics as smt
import torch
import torchmetrics as tm

from kit4dl.metric import MetricStore


class TestMetricStore:
    @pytest.fixture
    def metric_dict(self):
        yield {
            "Precision": tm.Precision(task="binary"),
            "Recall": tm.Recall(task="binary"),
        }

    @pytest.fixture
    def metric_store(self, metric_dict):
        yield MetricStore(metric_dict)

    def test_metricstore_results_repeatable(self, metric_store):
        metric_store.update(torch.LongTensor([0, 1]), torch.LongTensor([0, 1]))
        metric_store.update(torch.LongTensor([0, 1]), torch.LongTensor([1, 0]))
        res1 = metric_store.results
        res2 = metric_store.results
        assert id(res1) != id(res2)
        assert res1 == res2

    def test_get_empty_result_nan(self, metric_store):
        assert metric_store.results == {"Precision": np.nan, "Recall": np.nan}

    def test_reset_if_zeros_out(self, metric_store):
        metric_store.results == {"Precision": np.nan, "Recall": np.nan}
        metric_store.update(
            torch.randint(0, 2, size=(100,)), torch.randint(0, 2, size=(100,))
        )
        assert metric_store.results["Precision"] != 0.0
        assert metric_store.results["Recall"] != 0.0
        metric_store.reset()
        metric_store.results == {"Precision": np.nan, "Recall": np.nan}

    def test_metric_aggregation(self, metric_store):
        true = torch.randint(0, 2, size=(100,))
        true2 = torch.randint(0, 2, size=(100,))
        pred = torch.randint(0, 2, size=(100,))
        pred2 = torch.randint(0, 2, size=(100,))
        metric_store.update(true, pred)
        metric_store.update(true2, pred2)
        metrics_1 = metric_store.results

        metric_store.update(torch.cat([true, true2]), torch.cat([pred, pred2]))
        metrics_2 = metric_store.results
        assert metrics_1 == metrics_2

    def test_no_warning_on_before_update(self, metric_store):
        with warnings.catch_warnings(record=True) as warns:
            _ = metric_store.results
            assert len(warns) == 0



"""A module with predefined Kit4DL callbacks."""
import warnings

import lightning.pytorch as pl
from lightning.pytorch import callbacks as pl_callbacks

from kit4dl.nn.base import Kit4DLAbstractModule


class MetricCallback(pl_callbacks.Callback):
    """Callback that manages metrics.

    The callback resets metric stores on train or validation epoch start
    and logs metric on train or epoch end.
    """

    def on_train_epoch_start(
        self, trainer: pl.Trainer, pl_module: pl.LightningModule
    ) -> None:
        """Reset train metric store on start of each training stage."""
        if not isinstance(pl_module, Kit4DLAbstractModule):
            warnings.warn(
                f"type `{type(pl_module)}` doesn't support MetricCallback"
                " logic. try to inherit from the `kit4dl.Kit4DLAbstractModule`"
            )
            return
        pl_module.train_metric_tracker.reset()

    def on_train_epoch_end(
        self, trainer: pl.Trainer, pl_module: pl.LightningModule
    ) -> None:
        """Log metric values on the end of training stage."""
        if not isinstance(pl_module, Kit4DLAbstractModule):
            warnings.warn(
                f"type `{type(pl_module)}` doesn't support MetricCallback"
                " logic. try to inherit from the `kit4dl.Kit4DLAbstractModule`"
            )
            return
        pl_module.log_train_metrics()

    def on_validation_epoch_start(
        self, trainer: pl.Trainer, pl_module: pl.LightningModule
    ) -> None:
        """Reset validation metric store on start of each validation stage."""
        if not isinstance(pl_module, Kit4DLAbstractModule):
            warnings.warn(
                f"type `{type(pl_module)}` doesn't support MetricCallback"
                " logic. try to inherit from the `kit4dl.Kit4DLAbstractModule`"
            )
            return
        pl_module.val_metric_tracker.reset()

    def on_validation_epoch_end(
        self, trainer: pl.Trainer, pl_module: pl.LightningModule
    ) -> None:
        """Log metric values on the end of validation stage."""
        if not isinstance(pl_module, Kit4DLAbstractModule):
            warnings.warn(
                f"type `{type(pl_module)}` doesn't support MetricCallback"
                " logic. try to inherit from the `kit4dl.Kit4DLAbstractModule`"
            )
            return
        pl_module.log_val_metrics()

    def on_test_epoch_start(
        self, trainer: pl.Trainer, pl_module: pl.LightningModule
    ) -> None:
        """Reset test metric store on start of each test stage."""
        if not isinstance(pl_module, Kit4DLAbstractModule):
            warnings.warn(
                f"type `{type(pl_module)}` doesn't support MetricCallback"
                " logic. try to inherit from the `kit4dl.Kit4DLAbstractModule`"
            )
            return
        pl_module.test_metric_tracker.reset()

    def on_test_epoch_end(
        self, trainer: pl.Trainer, pl_module: pl.LightningModule
    ) -> None:
        """Log metric values on the end of test stage."""
        if not isinstance(pl_module, Kit4DLAbstractModule):
            warnings.warn(
                f"type `{type(pl_module)}` doesn't support MetricCallback"
                " logic. try to inherit from the `kit4dl.Kit4DLAbstractModule`"
            )
            return
        pl_module.log_test_metrics()


class ModelCheckpoint(pl_callbacks.ModelCheckpoint):
    """Callback for saving model checkpoint on fit end."""

    def on_fit_end(
        self, trainer: pl.Trainer, pl_module: pl.LightningModule
    ) -> None:
        """Save model weight on the end of training process."""
        super().on_fit_end(trainer, pl_module)



"""Module with session-related attributes."""
import sys
from typing import Any


class _ImmutableAttribute:
    _value: Any = None
    _default: Any = None

    def __init__(self, default: Any) -> None:
        self._default = default

    def __get__(self, obj, objtype=None) -> Any:
        return self._value if self._value else self._default

    def __set__(self, obj, value: Any) -> None:
        if self._value is not None:
            raise RuntimeError("Session properties can be set only once!")
        self._value = value


# NOTE: below definitions will not be used, they are required
# just for suntax suggestions and to avoid mypy [attr-defined] error
PROJECT_DIR: Any
LOG_LEVEL: Any
LOG_FORMAT: Any


def get_dict():
    """Get dictionary of all available context-defined properties."""


class Context:
    """Current Python session."""

    PROJECT_DIR: _ImmutableAttribute = _ImmutableAttribute(default=".")
    LOG_LEVEL: _ImmutableAttribute = _ImmutableAttribute(default="INFO")
    LOG_FORMAT: _ImmutableAttribute = _ImmutableAttribute(
        default="%(asctime)s - %(levelname)s - %(message)s"
    )

    def get_dict(self) -> dict:
        """Get dictionary of all available context-defined properties."""
        return {
            key: getattr(self, key)
            for key, value in type(self).__dict__.items()
            if isinstance(value, _ImmutableAttribute)
        }

    def __setattr__(self, name: str, value: Any) -> None:
        """
        Set the value for a context property.

        Parameters
        ----------
        name : str
            Name of the context property
        value : Any
            Value for the context property with the `key`

        Raises
        ------
        RuntimeError
            if the `name` is not defined (an attempt to create a new
            property)
        """
        if name not in Context.__dict__:
            # NOTE: maybe we can enable adding extra session arguments
            raise RuntimeError("Cannot set new session property!")
        super().__setattr__(name, value)


sys.modules[__name__] = Context()  # type: ignore[assignment]



"""Configuration formatting utils."""
import os

from jinja2 import (
    Environment,
    FileSystemLoader,
    StrictUndefined,
    UndefinedError,
)

_WINDOWS_PATHNAME_SEP = "\\"
_UNIX_PATHNAME_SEP = "/"


def substitute_symbols(path: str, **kwargs) -> str:
    """Replace placeholders with the provided keyword arguments and env vars.

    The function replaces `load` text with the provided keyword arguments
    and, eventually, all available environmental variables.

    Parameters
    ----------
    path : str
        Path to the file with placeholders
    **kwargs : Any
        Keword arguments to replace in `load`

    Returns
    -------
    replaced : str
        String with replaced placeholders

    Raises
    ------
    KeyError
        if a value was specified for any placeholder in `load`

    Examples
    --------
    ```python
    >>> load = "hi {{ name }}"
    >>> substitute_symbols(load, name="John")
    "hi John"
    ```

    ```python
    >>> load = "my env var has value {{ env['MY_ENV_VAR'] }}"
    >>> substitute_symbols(load)
    "my env var has value /usr/dataset"
    ```

    ```python
    >>> load = "hi {{ name }}"
    >>> substitute_symbols(load, age=10)
    KeyError no value found for the placeholder `name`
    ```
    """
    assert path is not None, "`path` cannot be `None`!"
    base_path, filename = os.path.split(path)
    tmpl_env = Environment(
        loader=FileSystemLoader(base_path), undefined=StrictUndefined
    )
    try:
        return tmpl_env.get_template(filename).render(env=os.environ, **kwargs)
    except UndefinedError as err:
        raise ValueError(err.message) from err


def escape_os_sep(load: str) -> str:
    """Escape OS-specific directory separator."""
    # NOTE: we make os separators uniform across the text load
    escaped = load.replace(_WINDOWS_PATHNAME_SEP, os.sep)
    escaped = escaped.replace(_UNIX_PATHNAME_SEP, os.sep)
    return escaped.replace(
        _WINDOWS_PATHNAME_SEP,
        "".join([_WINDOWS_PATHNAME_SEP, _WINDOWS_PATHNAME_SEP]),
    )



import pytest
import torch

from kit4dl.nn import validators
from tests.utils import skipnocuda


class TestValidators:
    def test_validate_cuda_device_exists_none(self):
        assert validators.validate_cuda_device_exists(None) is None

    @skipnocuda
    def test_validate_cuda_device_id_too_big(self):
        cuda_id = torch.cuda.device_count() + 1
        with pytest.raises(
            AssertionError,
            match=f"CUDA device with id `{cuda_id}` does not exist",
        ):
            validators.validate_cuda_device_exists(cuda_id)

    def test_validate_class_exists_fail_on_not_exising_qualified_path(self):
        with pytest.raises(ModuleNotFoundError, match=r"No module named 'a'"):
            validators.validate_class_exists("a.b::c")



from unittest.mock import patch

import pytest

try:
    import tomllib as toml
except ModuleNotFoundError:
    import toml  # type: ignore[no-redef]

from kit4dl.formatting import (
    _UNIX_PATHNAME_SEP,
    _WINDOWS_PATHNAME_SEP,
    escape_os_sep,
    substitute_symbols,
)


class TestFormatting:
    @pytest.fixture
    def load_path(self) -> str:
        import os
        from uuid import uuid4

        load = """
        [section]
        target = "{{ PROJECT_DIR }}/data.nc"
        """
        path = f"{uuid4()}.toml"
        with open(path, "wt") as file:
            file.writelines(load)
        yield path
        os.remove(path)

    @pytest.fixture
    def load_path_env(self) -> str:
        import os
        from uuid import uuid4

        load = """
        [section]
        target = "{{ env['env_path'] }}/data.nc"
        """
        path = f"{uuid4()}.toml"
        with open(path, "wt") as file:
            file.writelines(load)
        yield path
        os.remove(path)

    def test_substitute_symbols(self, load_path):
        result = substitute_symbols(load_path, PROJECT_DIR="/work/usr")
        entries = toml.loads(result)
        assert entries["section"]["target"] == "/work/usr/data.nc"

    def test_fail_on_missing_placeholder_definition(self, load_path):
        with pytest.raises(ValueError, match=r"'PROJECT_DIR' is undefined"):
            _ = substitute_symbols(load_path)

    @patch("os.sep", _UNIX_PATHNAME_SEP)
    def test_substitute_windows_path_to_unix(self):
        load = _WINDOWS_PATHNAME_SEP.join(
            ["D:", "Program Files", "some_file.txt"]
        )
        escaped = escape_os_sep(load)
        assert _WINDOWS_PATHNAME_SEP not in escaped
        assert _UNIX_PATHNAME_SEP in escaped

    @patch("os.sep", _WINDOWS_PATHNAME_SEP)
    def test_substitute_unix_path_to_windows(self):
        load = _UNIX_PATHNAME_SEP.join(
            ["D:", "Program Files", "some_file.txt"]
        )
        escaped = escape_os_sep(load)
        assert _UNIX_PATHNAME_SEP not in escaped
        assert _WINDOWS_PATHNAME_SEP in escaped

    @patch("os.sep", _WINDOWS_PATHNAME_SEP)
    def test_making_uniform_os_sep_windows(self):
        load = "D:\\Program Files/dir1/dir2\\file.txt"
        escaped = escape_os_sep(load)
        assert escaped == "D:\\\\Program Files\\\\dir1\\\\dir2\\\\file.txt"

    @patch("os.sep", _UNIX_PATHNAME_SEP)
    def test_making_uniform_os_sep_unix(self):
        load = "D:\\Program Files/dir1/dir2\\file.txt"
        escaped = escape_os_sep(load)
        assert escaped == "D:/Program Files/dir1/dir2/file.txt"

    def test_use_env_var_in_conf(self, load_path_env):
        import os

        os.environ["env_path"] = "/usr/new_path"
        result = substitute_symbols(load_path_env)
        entries = toml.loads(result)
        assert entries["section"]["target"] == "/usr/new_path/data.nc"



import os
import shutil

import pytest

from kit4dl.cli.app import init, train


@pytest.fixture
def experiment_name():
    return "new-project-1"


@pytest.fixture(autouse=True, scope="function")
def clear_new_project(experiment_name):
    shutil.rmtree(experiment_name, ignore_errors=True)
    yield
    shutil.rmtree(experiment_name, ignore_errors=True)


class TestApp:
    def test_init_creates_dir_with_name(self, experiment_name):
        init(name=experiment_name)
        assert os.path.exists(experiment_name)

    def test_init_creates_dir_default_name(self):
        init()
        assert os.path.exists("new_kit4dl_project")
        shutil.rmtree("new_kit4dl_project")

    def test_init_just_three_files_created(self, experiment_name):
        init(name=experiment_name)
        files = os.listdir(experiment_name)
        assert len(files) == 3
        assert "datamodule.py" in files
        assert "model.py" in files
        assert "conf.toml" in files



"""A module with neural network train task definition."""
import logging
from typing import Any

import lightning.pytorch as pl
from lightning.pytorch import callbacks as pl_callbacks
from lightning.pytorch import loggers as pl_log

from kit4dl.dataset import Kit4DLAbstractDataModule
from kit4dl.mixins import LoggerMixin
from kit4dl.nn.base import Kit4DLAbstractModule
from kit4dl.nn.callbacks import MetricCallback, ModelCheckpoint
from kit4dl.nn.confmodels import Conf
from kit4dl.utils import set_seed


class Trainer(LoggerMixin):
    """Class managing the training procedure."""

    _model: Kit4DLAbstractModule
    _datamodule: Kit4DLAbstractDataModule
    _trainer: pl.Trainer
    _conf: Conf
    _metric_logger: Any

    def __init__(self, conf: Conf) -> None:
        self._logger = logging.getLogger("lightning.pytorch")
        self._conf = conf
        self._device = self._conf.base.device
        self._metric_logger = self._new_metric_logger()
        set_seed(self._conf.base.seed)

    def prepare(self) -> "Trainer":
        """Prepare trainer by configuring the model and data modules."""
        self._model = self._configure_model()
        self._trainer = self._configure_trainer()
        self._datamodule = self._configure_datamodule()
        self._log_hparams()
        return self

    def _log_hparams(self) -> None:
        self._metric_logger.log_hyperparams(self._conf.dict())
        self._metric_logger.log_hyperparams(
            {
                "trainable_parameters": sum(
                    p.numel()
                    for p in self._model.parameters()
                    if p.requires_grad
                )
            }
        )

    def load_checkpoint(self, path: str) -> "Trainer":
        """Load model weights from the checkpoint."""
        self._model = type(self._model).load_from_checkpoint(path)
        return self

    def fit(self) -> "Trainer":
        """Fit the trainer making use of `lightning.pytorch.Trainer`."""
        assert self._trainer, (
            "trainer is not configured. did you forget to call `prepare()`"
            " method first?"
        )
        self._trainer.fit(self._model, datamodule=self._datamodule)
        return self

    def test(self) -> "Trainer":
        """Test the model."""
        assert self._trainer, (
            "trainer is not configured. did you forget to call `prepare()`"
            " method first?"
        )
        self._trainer.test(
            self._model, datamodule=self._datamodule, ckpt_path="best"
        )
        return self

    def predict(self) -> "Trainer":
        """Predict values for the model."""
        assert self._trainer, (
            "trainer is not configured. did you forget to call `prepare()`"
            " method first?"
        )
        self._trainer.predict(self._model, datamodule=self._datamodule)
        return self

    def _new_metric_logger(self) -> pl_log.Logger:
        return self._conf.logging.metric_logger_type(
            **self._conf.logging.arguments
        )

    def _configure_datamodule(self) -> Kit4DLAbstractDataModule:
        return self._conf.dataset.datamodule_class(conf=self._conf.dataset)

    def _configure_model(self) -> Kit4DLAbstractModule:
        return self._conf.model.model_class(conf=self._conf).to(self._device)

    def _get_model_checkpoint(self) -> ModelCheckpoint:
        assert self._conf.training.checkpoint, (
            "getting model checkpoint callback, but `checkpoint` was not"
            " defined in the configuration file"
        )
        chkp_conf = self._conf.training.checkpoint
        assert (not chkp_conf.every_n_epochs) or isinstance(
            chkp_conf.every_n_epochs, int
        ), (
            "wrong type of `every_n_epochs`. expected: `int`, provided:"
            f" {type(chkp_conf.every_n_epochs)}"
        )
        return ModelCheckpoint(
            dirpath=chkp_conf.path,
            filename=chkp_conf.filename,
            monitor=chkp_conf.monitor_metric_name,
            save_top_k=chkp_conf.save_top_k,
            mode=chkp_conf.mode,
            save_weights_only=chkp_conf.save_weights_only,
            every_n_epochs=chkp_conf.every_n_epochs,
            save_on_train_epoch_end=chkp_conf.save_on_train_epoch_end,
        )

    def _configure_trainer(self) -> pl.Trainer:
        accelerator_device, device = self._conf.base.accelerator_device_and_id
        callbacks: list[pl_callbacks.Callback] = [MetricCallback()]
        if self._conf.training.checkpoint:
            callbacks.append(self._get_model_checkpoint())
        return pl.Trainer(
            accelerator=accelerator_device,
            devices=device,
            max_epochs=self._conf.training.epochs,
            check_val_every_n_epoch=self._conf.validation.run_every_epoch,
            enable_progress_bar=True,
            deterministic=True,
            logger=self._metric_logger,
            callbacks=callbacks,
            **self._conf.training.arguments,
        )
